/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public enum EndingState { Escape, Entry, Lives, Credits, FlyAway }

    public string[] credits = {
      "STAFF",
      "",
      "",
      "",
      "",
      "Programmer",
      "    Michael Birken",
      "",
      "",
      "",
      "Graphic Designer",
      "    Michael Birken",
      "",
      "",
      "",
      "Editor",
      "    Michael Birken",
      "",
      "",
      "",
      "Planner",      
      "    Michael Birken",
      "",
      "",
      "",
      "Character Designer",
      "    Michael Birken",
      "",
      "",
      "",
      "Visual Designer",
      "    Michael Birken",
      "",
      "",
      "",
      "Object Designer",
      "    Michael Birken",
      "",
      "",
      "",
      "Director",
      "    Michael Birken",
      "",
      "",
      "",
      "Scroll Designer",
      "    Michael Birken",
      "",
      "",
      "",
      "Story",
      "    Michael Birken",
      "",
      "",
      "",
      "Technical Advisor",
      "    Michael Birken",
      "",
      "",
      "",
      "Code Guy",
      "    Michael Birken",
      "",
      "",
      "",
      "Special Thanks",
      "    Michael Birken",
      "",
      "",
      "",
      "Very Special Thanks",
      "    Michael Birken",
      "",
      "",
      "",
      "",
      "",
      "",
      "    and You!!",
      "",
      "",
      "",
      "",
      "",
      "",
      "",
      "",
      "",
      "SQUARIUS\u00B2",
      "Copyright \u00A9 2007 Michael Birken",
    };
    public int creditsY;
    public List<Star> stars = new List<Star>();
    public int endingX;
    public int endingY;
    public int endingVX;
    public int endingCX;
    public EndingState endingState;
    public double endingAngle;
    public double endingAngle2;
    public bool endingTitlesDone;
    public bool endingDone;
    public int endingLives;
    public string livesString;
    public int congradulationsX;
    public int congradulationsY;
    public int livesStringX;
    public int livesStringY;
    public int escapeX;
    public int escapeY;
    public int escapeXCenter;
    public double escapeXAngle;
    public Brush endingExplosionBrush;
    public bool clearedExplosion;
    public double escapeCenterVx;
    public double escapeCenterX;
    public int escapeFlash;
    public double escapeShipX;
    public double escapeShipVx;
    public string squaresDestroyedString;
    public int squaresDestroyedStringX;
    public int squaresDestroyedStringY;

    public void InitializeEnding() {
      endingState = EndingState.Escape;
      creditsY = 480;
      endingX = -64;
      endingVX = 1;
      endingAngle = endingAngle2 = 0;
      Text = "SQUARIUS\u00B2";
      endingTitlesDone = false;
      endingDone = false;
      for (int i = 0; i < 100; i++) {
        stars.Add(new Star(this, random.Next(640)));
      }
      endingLives = 640;
      livesString = (attempts < 2) ? "1 LIFE!  NO WAY!"
          : ("PLAYER LIVES: " + attempts);
      squaresDestroyedString = "OBLITERATED: " + squaresDestroyed;
      SizeF congradulationsSize = g.MeasureString("CONGRATULATIONS!!!", storyFont);
      SizeF livesStringSize = g.MeasureString(livesString, storyFont);
      SizeF squaresDestroyedSize = g.MeasureString(squaresDestroyedString, storyFont);
      congradulationsX = (int)((640 - congradulationsSize.Width) / 2);
      livesStringX = (int)((640 - livesStringSize.Width) / 2);
      squaresDestroyedStringX = (int)((640 - squaresDestroyedSize.Width) / 2);
      congradulationsY = 80;
      livesStringY = 300;
      squaresDestroyedStringY = (int)(livesStringY + livesStringSize.Height);
      escapeXCenter = 64;
      escapeXAngle = 0;
      explosions.Clear();
      ships.Clear();
      bullets.Clear();
      powerUps.Clear();
      endingExplosionBrush = Brushes.White;
      clearedExplosion = false;
      escapeCenterVx = 0;
      escapeCenterX = 0;
      escapeFlash = random.Next(40);
      escapeShipX = 0;
      escapeShipVx = 0;
    }

    public void UpdateModel_Ending() {
      endingAngle += 0.009;
      endingY = (int)(192 
          - 20 * Math.Sin(endingAngle) + 10 * Math.Sin(1.8 * endingAngle));
      switch (endingState) {
        case EndingState.Escape:
          if (clearedExplosion) {
            escapeXCenter = (int)escapeCenterX;
            escapeCenterX += escapeCenterVx;
            escapeCenterVx -= 0.1;
            escapeX = (int)(escapeShipX + 40 * Math.Sin(escapeXAngle));
            escapeShipX += escapeShipVx;
            if (escapeShipX > 1200) {
              endingState = EndingState.Entry;
              Text = "SQUARIUS\u00B2 - CONGRATULATIONS!!!";
            }
            if (escapeCenterX < -800) {
              escapeShipVx += 0.4;
            } else {
              escapeXAngle += 0.04;
            }
          } else {
            escapeXCenter++;
            escapeX = (int)(escapeXCenter + 40 * Math.Sin(escapeXAngle));
            if (escapeXCenter == 448) {
              clearedExplosion = true;
              escapeCenterX = 448;
              escapeShipX = 448;
              escapeShipVx = 0;
            }
            escapeXAngle += 0.04;  
          }                            
          escapeY = (int)(208 + 32 * Math.Sin(escapeXAngle * 1.7 + 0.3));
          if (escapeXCenter - 60 > 0) {
            explosions.Add(new Explosion(this,
                random.Next(escapeXCenter - 60), random.Next(480), 100));            
          }
          for (int i = explosions.Count - 1; i >= 0; i--) {
            Explosion explosion = explosions[i];
            if (explosion.Move()) {
              explosions.RemoveAt(i);
            }
          }
          endingExplosionBrush = (endingExplosionBrush == Brushes.White)
              ? Brushes.Yellow : Brushes.White;
          foreach (Star star in stars) {
            star.Move();
          }
          foreach (Star star in stars) {
            star.Move();
          }
          foreach (Star star in stars) {
            star.Move();
          }
          if (escapeFlash > 0) {
            escapeFlash--;
          } else {
            escapeFlash = random.Next(40);
          }
          break;
        case EndingState.Entry:
          endingX += endingVX;
          if (endingX >= 350) {
            endingState = EndingState.Lives;
          }
          break;
        case EndingState.Lives:
          endingAngle2 += 0.004;
          endingX = (int)(350 + 50 * Math.Sin(1.6 * endingAngle2)
              + 25 * Math.Sin(3.1 * endingAngle2));
          endingLives--;
          if (endingLives == 0) {
            endingState = EndingState.Credits;
          }
          break;
        case EndingState.Credits:
          endingAngle2 += 0.004;
          endingX = (int)(350 + 50 * Math.Sin(1.6 * endingAngle2) 
              + 25 * Math.Sin(3.1 * endingAngle2));
          creditsY--;
          if (endingTitlesDone) {
            endingState = EndingState.FlyAway;
          }
          break;
        case EndingState.FlyAway:
          if (endingX < 640) {
            endingVX += 1;
          } else {
            endingVX = 1;
          }
          endingX += endingVX;
          if (endingX > 750) {
            endingDone = true;
          }
          break;
      }      
      foreach (Star star in stars) {
        star.Move();
      }
    }

    public void RenderFrame_Ending() {

      if (endingState == EndingState.Escape) {
        g.FillRectangle(escapeFlash == 0 && !clearedExplosion 
            ? Brushes.White : Brushes.Black, 
            0, 0, 640, 480);
        foreach (Star star in stars) {
          g.DrawLine(Pens.White, star.x, star.y, star.x + 60, star.y);
        }
        if (escapeXCenter - 32 > 0) {
          g.FillEllipse(endingExplosionBrush, escapeXCenter - 128, 0, 128, 480);
          g.FillRectangle(endingExplosionBrush, 0, 0, escapeXCenter - 64, 480);
        }
        g.FillRectangle(Brushes.Red, escapeX, escapeY, 64, 64);
        foreach (Explosion explosion in explosions) {
          g.FillEllipse(explosion.brush, explosion.x - explosion.radius,
              explosion.y - explosion.radius, explosion.radius << 1,
              explosion.radius << 1);
        }
        return;
      }

      g.FillRectangle(Brushes.Black, 0, 0, 640, 480);
      foreach(Star star in stars) {
        g.FillRectangle(star.brush, star.x, star.y, 1, 1);
      }

      g.FillRectangle(Brushes.Red, endingX, endingY, 64, 64);

      if (endingState == EndingState.Lives) {
        g.DrawString("CONGRATULATIONS!!!", storyFont, Brushes.Yellow,
            congradulationsX, congradulationsY);
        g.DrawString(livesString, storyFont, Brushes.White,
            livesStringX, livesStringY);
        g.DrawString(squaresDestroyedString, storyFont, Brushes.White,
            squaresDestroyedStringX, squaresDestroyedStringY);
      } else if (endingState == EndingState.Credits) {
        int j = creditsY;
        for (int i = 0; i < credits.Length;
            i++, j += subtitleFontHeight) {
          if (j < 490 && j > -subtitleFontHeight - 10 && credits[i].Length > 0) {
            g.DrawString(credits[i], subtitleFont,
                credits[i][0] == ' ' ? Brushes.White : Brushes.Yellow, 50, j);
          }
        }
        if (j < -subtitleFontHeight - 50) {
          endingTitlesDone = true;
        }
      }

      if (endingDone) {
        stars.Clear();
        explosions.Clear();
        StartIntroduction();
      }
    }
  }

  public class Star {
    public int x;
    public int y;
    public int vx;
    public Brush brush;
    public SquariusForm squariusForm;

    public Star(SquariusForm squariusForm, int x) {
      this.squariusForm = squariusForm;
      this.x = x;
      Reset();
    }

    public void Reset() {
      y = squariusForm.random.Next(480);
      vx = -1 - squariusForm.random.Next(16);
      switch (squariusForm.random.Next(3)) {
        case 0:
          brush = Brushes.PaleTurquoise;
          break;
        case 1:
          brush = Brushes.AliceBlue;
          break;
        default:
          brush = Brushes.PowderBlue;
          break;
      }
    }

    public void Move() {
      x += vx;
      if (x < 0) {
        x = 640;
        Reset();
      }
    }
  }
}