/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public enum IntroState { StartUpDelay, OneShip, Armada, Rising, Blockhead,
        Lucky, Rolling, TakeOff, DropTitle, FlashEnter };
    public string[] introStrings = {
      "One ship",
      "against an entire armada...",
      "Only a blockhead would",
      "take on this mission...",
      "Lucky they found me!",
    };
    public string[] controlStrings = {
      "D - Hold down to shoot",
      "Arrow Keys  Move",
      "P  Pause",
      "Backspace  Restart Stage",
      "F11 - Toggle Full Screen Mode",
      "\u00A9 1979 Squarenami Consortium, Ltd.",
      "PRESS ENTER",
    };
    public int[] introStringX;
    public int[] controlStringX;

    public Font storyFont;
    public Font titleFont;
    public Font subtitleFont;
    public int introDelay;
    public IntroState introState;
    public int introY;
    public double introX;
    public double introVX;
    public double introVY;
    public int roadX;
    public double liftY;
    public SizeF titleSize;
    public SizeF subtitleSize;
    public int titleX;
    public int subtitleX;
    public int enterCountDown;
    public int storyFontHeight;
    public int subtitleFontHeight;
    public int titleFontHeight;

    public void InitializeIntroduction() {
      introState = IntroState.StartUpDelay;
      introDelay = 40;
      if (storyFont == null) {
        storyFont = new Font(Font.FontFamily.Name, 32, FontStyle.Bold);
        titleFont = new Font(Font.FontFamily.Name, 64, FontStyle.Bold);
        subtitleFont = new Font(Font.FontFamily.Name, 16, FontStyle.Bold);
        storyFontHeight = storyFont.Height;
        titleFontHeight = titleFont.Height;
        subtitleFontHeight = subtitleFont.Height;
        titleSize = g.MeasureString("SQUARIUS\u00B2", titleFont);        
        titleX = (int)((640 - titleSize.Width) / 2);
        subtitleSize = g.MeasureString("- S Q U A R E D -", subtitleFont);
        subtitleX = (int)((640 - subtitleSize.Width) / 2);
        introStringX = new int[introStrings.Length];
        controlStringX = new int[controlStrings.Length];
        for (int i = 0; i < introStrings.Length; i++) {
          introStringX[i] = (int)((640
              - g.MeasureString(introStrings[i], storyFont).Width) / 2);          
        }
        for (int i = 0; i < controlStrings.Length; i++) {
          controlStringX[i] = (int)((640
              - g.MeasureString(controlStrings[i], subtitleFont).Width) / 2);
        }
      }
      Text = "SQUARIUS\u00B2";
    }

    public void UpdateModel_Introduction() {
      switch (introState) {
        case IntroState.StartUpDelay:
          if (introDelay > 0) {
            introDelay--;
          } else {
            introDelay = 50;
            introState = IntroState.OneShip;
          }
          break;
        case IntroState.OneShip:
          if (introDelay > 0) {
            introDelay--;
          } else {
            introDelay = 140;
            introState = IntroState.Armada;
          }
          break;
        case IntroState.Armada:
          if (introDelay > 0) {
            introDelay--;
          } else {
            introY = 480;
            introDelay = 20;
            introState = IntroState.Rising;
          }
          break;
        case IntroState.Rising:
          if (introY > 124) {
            introY -= 2;
          } else {
            if (introDelay > 0) {
              introDelay--;
            } else {
              introDelay = 140;
              introState = IntroState.Blockhead;
            }
          }
          break;
        case IntroState.Blockhead:
          if (introDelay > 0) {
            introDelay--;
          } else {
            introDelay = 140;
            introState = IntroState.Lucky;
            introX = 64;
            introVX = 1;
          }
          break;
        case IntroState.Lucky:
          if (introDelay > 0) {
            introDelay--;
          } else {
            introDelay = 30;
            introState = IntroState.Rolling;
          }
          break;
        case IntroState.Rolling:
          if (introDelay > 0) {
            introDelay--;
          } else {
            if (introX < 320) {
              introX += introVX;
              introVX += 0.4;
            } else {
              introState = IntroState.TakeOff;
              introX = 64;
              roadX = 0;
              liftY = 416;
              introVX = 1;
              introVY = 0;
            }
          }
          break;
        case IntroState.TakeOff:
          roadX -= 16;
          introX += introVX;
          if (introX > 680) {
            introState = IntroState.DropTitle;
            introX = -128 - 32 - titleSize.Width - titleX;
          }
          if (introX > 128) {
            liftY += introVY;
            if (liftY < 250) {
              if (introVY <= 0) {
                introVY += 0.1;
                if (introVY >= 0) {
                  introVY = 0;                  
                }
                if (introVY == 0) {
                  introVX += 0.6;
                }
              } 
            } else {
              introVY -= 0.05;
            }
          }
          if (roadX < -192) {
            roadX = 0;
          }
          break;
        case IntroState.DropTitle:
          introX += 32;
          if (introX > 640) {
            introState = IntroState.FlashEnter;
          }
          enterCountDown = 600;
          break;
        case IntroState.FlashEnter:
          enterCountDown--;
          if (enterCountDown == 0) {
            InitializeIntroduction();
          }
          break;
      }
    }

    public void RenderFrame_Introduction() {

      g.FillRectangle(Brushes.Black, 0, 0, 640, 480);

      switch (introState) {
        case IntroState.StartUpDelay:
          break;
        case IntroState.OneShip:
          g.DrawString(introStrings[0], storyFont, Brushes.LightSalmon,
              introStringX[0], 100);
          break;
        case IntroState.Armada:
          g.DrawString(introStrings[0], storyFont, Brushes.LightSalmon,
              introStringX[0], 100);
          g.DrawString(introStrings[1], storyFont, Brushes.LightSalmon,
              introStringX[1], 100 + storyFontHeight);
          break;
        case IntroState.Rising:
          g.FillRectangle(Brushes.Silver, 0, 380, 32, 100);
          g.FillRectangle(Brushes.Silver, 352, 380, 288, 100);
          g.FillRectangle(Brushes.Red, 64, introY, 256, 256);
          g.FillRectangle(Brushes.Silver, 32, introY + 256, 320, 100);
          break;
        case IntroState.Blockhead:
          g.DrawString(introStrings[2], storyFont, Brushes.LightSalmon,
              introStringX[2], 100);
          g.DrawString(introStrings[3], storyFont, Brushes.LightSalmon,
              introStringX[3], 100 + storyFontHeight);
          break;
        case IntroState.Lucky:
          g.DrawString(introStrings[2], storyFont, Brushes.LightSalmon,
              introStringX[2], 100);
          g.DrawString(introStrings[3], storyFont, Brushes.LightSalmon,
              introStringX[3], 100 + storyFontHeight);
          g.DrawString(introStrings[4], storyFont, Brushes.LightSalmon,
                        introStringX[4], 100 + storyFontHeight * 3);
          break;
        case IntroState.Rolling:
          g.FillRectangle(Brushes.Silver, 0, 380, 640, 100);
          g.FillRectangle(Brushes.Red, (int)introX, 124, 256, 256);
          break;
        case IntroState.TakeOff:
          for (int i = roadX, j = 0; i <= 736; i += 96, j++) {
            g.FillRectangle((j & 1) == 0 ? Brushes.Silver : Brushes.Gray,
                i, 448, 96, 32);
          }
          g.FillRectangle(Brushes.Red, (int)introX, (int)liftY, 32, 32);
          break;
        case IntroState.DropTitle: {
            int _titleX = (int)Math.Min(0, introX);
            g.DrawString("SQUARIUS\u00B2", titleFont, Brushes.Red, 
              _titleX + titleX, 70);
            g.DrawString("- S Q U A R E D -", subtitleFont,
                Brushes.Red, _titleX + subtitleX, 70 + titleFontHeight);
            g.FillRectangle(Brushes.Red, (int)introX + titleX + titleSize.Width,
                70, 128, 128);
            break;
          }
        case IntroState.FlashEnter: {
          g.DrawString("SQUARIUS\u00B2", titleFont, Brushes.Red, titleX, 70);
          g.DrawString("- S Q U A R E D -", subtitleFont,
              Brushes.Red, subtitleX, 70 + titleFontHeight);
          int j = (int)(100 + titleFontHeight + subtitleSize.Height);
          for (int i = 0; i < 5; i++, j += (int)subtitleSize.Height) {
            g.DrawString(controlStrings[i], subtitleFont,
              Brushes.Yellow, controlStringX[i], j);
          }
          int copyrightY = (int)(465 - subtitleSize.Height);
          g.DrawString(controlStrings[5], subtitleFont,
              Brushes.Yellow, controlStringX[5], copyrightY);
          if (((enterCountDown >> 4) & 1) == 0) {
            g.DrawString(controlStrings[6], subtitleFont,
                Brushes.White, controlStringX[6], 
                (j + copyrightY - (int)subtitleSize.Height) >> 1);
          }
          break;
        }
      }

      if (enterKeyPressed) {
        StartPlaying();                        
      }
    }
  }
}