/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public List<Ship> ships = new List<Ship>();
    public List<Explosion> explosions = new List<Explosion>();
    public List<Bullet> bullets = new List<Bullet>();
    public List<PowerUp> powerUps = new List<PowerUp>();
    public Ship playerShip;
    public bool playerAlive = true;
    public int weaponLevel = 0;
    public int shootDelay;
    public int[,] seekerPoints = new int[20, 2];
    public int seekerIndex;
    public int restorePlayerCount;
    public int squaresDestroyed;
    public int checkPointViewportX;
    public int checkPointWeaponLevel;
    public int checkPointScrollSpeed = 2;
    public int checkPointSquaresDestroyed;
    public int weaponsCollectedOnStage;
    public int fireworks;
    public bool requestAdvanceStage;
    public bool requestRestartStage;

    private void UpdateModel_Playing() {

      if (requestAdvanceStage) {
        requestAdvanceStage = false;
        AdvanceStage();
      } else if (requestRestartStage) {
        requestRestartStage = false;
        RestartStage();
      } else if (fireworks > 0) {
        fireworks--;
        if ((fireworks & 3) == 0) {
          explosions.Add(new Explosion(this, 
              viewportX + random.Next(640), random.Next(480)));
        }
        if (fireworks == 0) {
          playerShip.x -= viewportX;
          if (stage + 1 < STAGES) {
            stage++;
            InitStage();
          } else {
            StartEnding();
          }          
        }
      }

      if ((viewportX & 0x1F) == 0 && viewportX > lastViewPortTrigger) {
        lastViewPortTrigger = viewportX;
        int X = viewportX + 640;
        for (int i = 0, y = 0; i < 15; i++, y += 32) {
          ProcessCell(GetCell((viewportX >> 5) + 20, i), X, y);
        }

/*#if DEBUG
        Console.WriteLine("Ships: {0}", ships.Count);
        Console.WriteLine("Bullets: {0}", bullets.Count);
        Console.WriteLine("Explosions: {0}", explosions.Count);
#endif*/
      }
      viewportX += scrollSpeed;

      if (playerAlive) {
        if (shootDelay > 0) {
          shootDelay--;
        }

        playerShip.x += scrollSpeed;

        if (leftKeyPressed) {
          playerShip.x -= 8;
          if (playerShip.x < viewportX) {
            playerShip.x = viewportX;
          }
        } else if (rightKeyPressed) {
          playerShip.x += 8;
          if (playerShip.x >= 608 + viewportX) {
            playerShip.x = 608 + viewportX;
          }
        }
        if (upKeyPressed) {
          playerShip.y -= 8;
          if (playerShip.y < 0) {
            playerShip.y = 0;
          }
        } else if (downKeyPressed) {
          playerShip.y += 8;
          if (playerShip.y >= 448) {
            playerShip.y = 448;
          }
        }
        if (shootKeyPressed) {
          PlayerShoot();
        }

        if (fireworks == 0 && (IsWall(playerShip.x, playerShip.y)
            || IsWall(playerShip.x + playerShip.width - 1, playerShip.y)
            || IsWall(playerShip.x, playerShip.y + playerShip.height - 1)
            || IsWall(playerShip.x + playerShip.width - 1,
                playerShip.y + playerShip.height - 1))) {          
          PlayerKilled();
        }

        seekerPoints[seekerIndex, 0] = playerShip.x - 32;
        seekerPoints[seekerIndex, 1] = playerShip.y + 16;
        seekerIndex--;
        if (seekerIndex < 0) {
          seekerIndex = 19;
        }
      } else if (--restorePlayerCount == 0) {
        RestorePlayer();
      }

      for (int i = powerUps.Count - 1; i >= 0; i--) {
        PowerUp powerUp = powerUps[i];
        if (powerUp.Move()) {
          powerUps.RemoveAt(i);
        }
      }

      for (int i = ships.Count - 1; i >= 0; i--) {
        Ship ship = ships[i];
        if (ship.Move()) {
          ships.RemoveAt(i);
        } else if (playerAlive && ship.Intersects(playerShip)) {
          PlayerKilled();
        }
      }

      for (int i = bullets.Count - 1; i >= 0; i--) {
        Bullet bullet = bullets[i];
        if (bullet.Move()) {
          bullets.RemoveAt(i);
        }
      }

      for (int i = explosions.Count - 1; i >= 0; i--) {
        Explosion explosion = explosions[i];
        if (explosion.Move()) {
          explosions.RemoveAt(i);
        }
      }
    }

    private void RenderFrame_Playing() {

      g.FillRectangle(Brushes.Black, 0, 0, 640, 480);

      foreach (Bullet bullet in bullets) {
        g.FillEllipse(bullet.Brush, bullet.X - viewportX, bullet.Y, 8, 8);
      }

      foreach (Ship ship in ships) {
        g.FillRectangle(ship.brush, ship.x - viewportX, ship.y,
            ship.width, ship.height);
      }

      RenderWalls();

      if (playerAlive) {
        g.FillRectangle(playerShip.brush, playerShip.x - viewportX, playerShip.y,
            playerShip.width, playerShip.height);
        if (weaponLevel > 4) {
          g.FillEllipse(playerShip.brush, 
              seekerPoints[seekerIndex, 0] - 10 - viewportX,
              seekerPoints[seekerIndex, 1] - 10, 20, 20);
        }
        if (weaponLevel > 5) {
          int shields = weaponLevel - 5;
          for (int i = 0, j = 8, k = 15; i < shields; i++, j++, k += 2) {
            g.DrawEllipse(Pens.LightGreen,
                playerShip.x - j - viewportX, playerShip.y - j,
                playerShip.width + k, playerShip.height + k);
          }
        }
      }

      foreach (PowerUp powerUp in powerUps) {
        g.FillRectangle(
            powerUp.brush, powerUp.x - viewportX, powerUp.y, 
            powerUp.width, powerUp.height);
      }

      foreach (Explosion explosion in explosions) {
        g.FillEllipse(explosion.brush, explosion.x - explosion.radius - viewportX,
            explosion.y - explosion.radius, explosion.radius << 1,
            explosion.radius << 1);
      }
    }

    private void RenderWalls() {
      if (fireworks > 0) {
        return;
      }
      int startCol = viewportX >> 5;
      int screenX = -(viewportX & 0x1F);
      for (int row = 0, y = 0; row < 15; row++, y += 32) {
        for (int col = startCol, x = screenX; x < 640; col++, x += 32) {
          if (GetCell(col, row) == 'X') {
            g.FillRectangle(wallBrush, x, y, 32, 32);
          }
        }
      }
    }

    public char GetCell(int x, int y) {
      if (x < 0 || y < 0 || y >= 15 || x >= stages[stage, y].Length) {
        return '.';
      }
      return stages[stage, y][x];
    }

    public bool IsWall(int x, int y) {
      int X = x >> 5;
      int Y = y >> 5;
      if (X < 0 || Y < 0 || Y >= 15 || X >= stages[stage, Y].Length) {
        return false;
      }
      return stages[stage, Y][X] == 'X';
    }

    private void PlayerShoot() {
      if (shootDelay > 0) {
        return;
      }
      shootDelay = 15;
      int halfY = playerShip.y + (playerShip.height >> 1);
      if (weaponLevel == 0) {
        for (int i = 0; i < 3; i++) {
          bullets.Add(new Bullet(this, BulletType.Player,
              playerShip.x + playerShip.width + 10 * i - 20, halfY, 
              18 + scrollSpeed, 0));
        }
      } else if (weaponLevel == 1) {
        for (int i = 0; i < 4; i++) {
          int x = playerShip.x + playerShip.width + 10 * i - 20;
          bullets.Add(new Bullet(this, BulletType.Player,
              x, playerShip.y + 4, 18 + scrollSpeed, 0));
          bullets.Add(new Bullet(this, BulletType.Player,
              x, playerShip.y + 27, 18 + scrollSpeed, 0));
        }
      } else if (weaponLevel > 1) {
        int[,] bulletMap = bulletMaps[weaponLevel == 2 ? 0 : 1];
        for (int i = bulletMap.GetLength(0) - 1; i >= 0; i--) {
          bullets.Add(new Bullet(this, BulletType.Player,
            playerShip.x + bulletMap[i, 0],
            playerShip.y + bulletMap[i, 1],
            18 + scrollSpeed, 0));
        }
      } 
      if (weaponLevel > 3) {
        for (int i = 0; i < 3; i++) {
          int x = playerShip.x + playerShip.width + 7 * i - 20;
          bullets.Add(new Bullet(this, BulletType.Player,
              x, playerShip.y + 4 - 7 * i, 12.14 + scrollSpeed, -14.14));
          bullets.Add(new Bullet(this, BulletType.Player,
              x, playerShip.y + 27 + 7 * i, 12.14 + scrollSpeed, 14.14));
        }
      }
      if (weaponLevel > 4) {        
        for (int i = 0; i < 3; i++) {
          bullets.Add(new Bullet(this, BulletType.Player,
              seekerPoints[seekerIndex, 0] + 10 * i - 8,
              seekerPoints[seekerIndex, 1], 18 + scrollSpeed, 0));
        }
      }
    }

    private void ResetSeeker() {
      for(int i = seekerPoints.GetLength(0) - 1; i >= 0; i--) {
        seekerPoints[i, 0] = playerShip.x - playerShip.width;
        seekerPoints[i, 1] = playerShip.y + (playerShip.height >> 1);
      }
    }

    public void PlayerKilled() {
      explosions.Add(new Explosion(this, playerShip));
      playerAlive = false;
      restorePlayerCount = 80;
      attempts++;
    }

    public void RestorePlayer() {      
/*#if DEBUG
      checkPointViewportX = 34592;
      checkPointWeaponLevel = 8;
      checkPointScrollSpeed = 2;
#endif*/

      playerAlive = true;
      lastViewPortTrigger = -1;
      viewportX = checkPointViewportX;
      playerShip.x = 96 + viewportX;
      playerShip.y = 224;
      weaponsCollectedOnStage = weaponLevel = checkPointWeaponLevel;
      scrollSpeed = checkPointScrollSpeed;
      squaresDestroyed = checkPointSquaresDestroyed;
      fireworks = 0;
      requestAdvanceStage = false;
      requestRestartStage = false;
      ResetSeeker();
      bullets.Clear();
      powerUps.Clear();
      ships.Clear();
      explosions.Clear();
    }

    private void ProcessCell(char cell, int x, int y) {
      switch (stage) {
        case 0:
          ProcessCell_0(cell, x, y);
          break;
        case 1:
          ProcessCell_1(cell, x, y);
          break;
        case 2:
          ProcessCell_2(cell, x, y);
          break;
        case 3:
          ProcessCell_3(cell, x, y);
          break;
        case 4:
          ProcessCell_4(cell, x, y);
          break;
        case 5:
          ProcessCell_5(cell, x, y);
          break;
        case 6:
          ProcessCell_6(cell, x, y);
          break;
        case 7:
          ProcessCell_7(cell, x, y);
          break;
      }
    }

    private void InitStage() {
      switch (stage) {
        case 0:
          InitStage_0();
          break;
        case 1:
          InitStage_1();
          break;
        case 2:
          InitStage_2();
          break;
        case 3:
          InitStage_3();
          break;
        case 4:
          InitStage_4();
          break;
        case 5:
          InitStage_5();
          break;
        case 6:
          InitStage_6();
          break;
        case 7:
          InitStage_7();
          break;
      }      
      viewportX = 0;
      lastViewPortTrigger = -1;
      checkPointViewportX = 0;
      checkPointWeaponLevel = 0;
      weaponsCollectedOnStage = 0;
      checkPointScrollSpeed = scrollSpeed;
    }

    public void MakeCheckPoint() {
      checkPointViewportX = viewportX;
      checkPointWeaponLevel = weaponsCollectedOnStage;
      checkPointScrollSpeed = scrollSpeed;
      checkPointSquaresDestroyed = squaresDestroyed;

/*#if DEBUG
      Console.WriteLine("checkPointViewportX = {0};", checkPointViewportX);
      Console.WriteLine("checkPointWeaponLevel = {0};", checkPointWeaponLevel);
      Console.WriteLine("checkPointScrollSpeed = {0};", checkPointScrollSpeed);
      Console.WriteLine("checkPointSquaresDestroyed = {0};", checkPointSquaresDestroyed);
#endif*/
    }

    public void AdvanceStage() {
      ships.Clear();
      for(int i = bullets.Count - 1; i >= 0; i--) {
        Bullet bullet = bullets[i];
        if (bullet.bulletType == BulletType.Enemy) {
          bullets.RemoveAt(i);
        }
      }
      fireworks = 200;
    }
  }
}
