/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing;

namespace Squarius2 {

  [StructLayout(LayoutKind.Sequential)]
  public struct DeviceMode {
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
    public string deviceName;
    public short specVersion;
    public short driverVersion;
    public short size;
    public short driverExtra;
    public int fields;

    public short orientation;
    public short paperSize;
    public short paperLength;
    public short paperWidth;

    public short scale;
    public short copies;
    public short defaultSource;
    public short printQuality;
    public short color;
    public short duplex;
    public short yResolution;
    public short ttOption;
    public short collate;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
    public string formName;
    public short logPixels;
    public short bitsPerPel;
    public int pelsWidth;
    public int pelsHeight;

    public int displayFlags;
    public int displayFrequency;

    public int icmMethod;
    public int icmIntent;
    public int mediaType;
    public int ditherType;
    public int reserved1;
    public int reserved2;

    public int panningWidth;
    public int panningHeight;
  };

  public class ResolutionManager {

    [DllImport("user32.dll")]
    public static extern int EnumDisplaySettings(string deviceName, int modeNum, 
        ref DeviceMode devMode);
    [DllImport("user32.dll")]
    public static extern int ChangeDisplaySettings(ref DeviceMode devMode, int flags);

    public const int ENUM_CURRENT_SETTINGS = -1;
    public const int CDS_UPDATEREGISTRY = 0x01;
    public const int CDS_TEST = 0x02;
    public const int DISP_CHANGE_SUCCESSFUL = 0;
    public const int DISP_CHANGE_RESTART = 1;
    public const int DISP_CHANGE_FAILED = -1;

    public static Rectangle screenBounds;
    public static Point lastLocation;
    public static FormBorderStyle borderStyle;

    public static void EnterFullScreenMode(Form form, int width, int height) {
      screenBounds = Screen.PrimaryScreen.Bounds;
      lastLocation = form.Location;
      borderStyle = form.FormBorderStyle;
      form.FormBorderStyle = FormBorderStyle.None;
      form.Location = new Point(0, 0);
      form.TopMost = true;
      ChangeScreenResolution(width, height);
    }

    public static void ExitFullScreenMode(Form form) {
      ChangeScreenResolution(screenBounds.Width, screenBounds.Height);
      form.Location = lastLocation;
      form.FormBorderStyle = borderStyle;
      form.TopMost = false;
    }
    
    public static void ChangeScreenResolution(int width, int height) {
      DeviceMode deviceMode = new DeviceMode();
      deviceMode.deviceName = new String(new char[32]);
      deviceMode.formName = new String(new char[32]);
      deviceMode.size = (short)Marshal.SizeOf(deviceMode);

      if (EnumDisplaySettings(null, ENUM_CURRENT_SETTINGS, ref deviceMode) != 0) {

        deviceMode.pelsWidth = width;
        deviceMode.pelsHeight = height;

        if (ChangeDisplaySettings(ref deviceMode, CDS_TEST) != DISP_CHANGE_FAILED) {
          ChangeDisplaySettings(ref deviceMode, CDS_UPDATEREGISTRY);
        }
      }
    }
  }
}

