/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_0() {
      scrollSpeed = 2;
      wallBrush = Brushes.SandyBrown;
      Text = "SQUARIUS\u00B2 - Stage 1";      
    }

    public void ProcessCell_0(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'B':
          RightAngleShip.CreateFleet(this, x, y, false);
          break;
        case 'C':
          RightAngleShip.CreateFleet(this, x, y, true);
          break;
        case 'D':
          ships.Add(new Creeper(this, x, y, true, false));
          break;
        case 'E':
          ships.Add(new Creeper(this, x, y, false, false));
          break;
        case 'F':
          HorizontalShip.CreateFleet(this, x, y, false, false);
          break;
        case 'G':
          HorizontalShip.CreateFleet(this, x, y, true, false);
          break;
        case 'H':
          HorizontalShip.CreateFleet(this, x, y, false, true);
          break;
        case 'I':
          HorizontalShip.CreateFleet(this, x, y, true, true);
          break;
        case 'J':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'K':
          ships.Add(new Stalagtite(this, x, y, true));
          break;
        case 'L':
          ships.Add(new Stalagtite(this, x, y, false));
          break;
        case 'Y':
          scrollSpeed = 0;
          break;
        case 'Z':
          ships.Add(new PongBoss(this, x));
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class HorizontalShip : Ship {

    public int shootCounter = 10;
    public bool shoots;

    public HorizontalShip(SquariusForm squariusForm, int x, int y,
        bool yieldsPowerUp, bool shoots) : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.yieldsPowerUp = yieldsPowerUp;
      this.shoots = shoots;

      if (yieldsPowerUp) {
        brush = Brushes.Orange;
      } else if (shoots) {
        brush = Brushes.Magenta;
      } else {
        brush = Brushes.DodgerBlue;
      }
    }

    public override bool Move() {
      if (shoots) {
        if (shootCounter == 0) {
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
          shootCounter = 80;
        } else {
          shootCounter--;
        }
      }
      x -= 2;
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return true;
    }

    public static void CreateFleet(SquariusForm squariusForm, int x, int y,
        bool yieldsPowerUp, bool shoots) {
      squariusForm.ships.Add(new HorizontalShip(
          squariusForm, x, y, yieldsPowerUp, shoots));
      squariusForm.ships.Add(new HorizontalShip(
          squariusForm, x, y - 64, false, shoots));
      squariusForm.ships.Add(new HorizontalShip(
          squariusForm, x, y + 64, false, shoots));
    }
  }

  public class WallGun : Ship {

    public int shootCounter = 10;

    public WallGun(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp) 
        : base(squariusForm) {
      brush = yieldsPowerUp ? Brushes.Orange : Brushes.RosyBrown;
      this.x = x;
      this.y = y;
      this.yieldsPowerUp = yieldsPowerUp;
    }

    public override bool Move() {
      if (shootCounter == 0) {
        squariusForm.bullets.Add(new Bullet(squariusForm, this));
        shootCounter = 80;
      } else {
        shootCounter--;
      }
      return x + width < squariusForm.viewportX;      
    }

    public override bool Shot() {
      return true;
    }
  }

  public class Ball : Ship {

    public int shootCounter = 10;
    public int flashing;
    public int vx = 2;
    public int vy = -3;
    public int X;
    public int hits;
    public Brush defaultBrush;
    public int entryDelay = 10;

    public Ball(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      defaultBrush = brush = Brushes.MediumPurple;
      this.x = x;
      X = x - 32;
      this.y = y;
      width = height = 0;
    }

    public override bool Move() {
      if (width < 64) {
        width += 4;
        height += 4;
        x -= 2;
        y -= 2;
        return false;
      } else if (entryDelay > 0) {
        entryDelay--;
        return false;
      }
      if (shootCounter == 0) {
        double shootAngle = squariusForm.random.NextDouble() * Math.PI * 2.0;
        double sin = Math.Sin(shootAngle);
        double cos = Math.Cos(shootAngle);
        double vx = 6 * cos;
        double vy = 6 * sin;
        for (int i = 0, radius = 0; i < 7; i++, radius += 8) {
          squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
              x + 32 + radius * cos, y + 32 + radius * sin, 
              vx, vy));
        }
        shootCounter = 32;
      } else {
        shootCounter--;
      }
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
        }
      } 
      if (y <= 32) {
        vy = 3;
      } else if (y >= 384) {
        vy = -3;
      }
      if (x <= X) {
        vx = 2;
      } else if (x >= X + 448) {
        vx = -2;
      }
      x += vx;
      y += vy;
      return false;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        brush = Brushes.White;
        hits++;
        if (hits == 80) {
          brush = defaultBrush = Brushes.Magenta;
        } else if (hits == 100) {
          squariusForm.requestAdvanceStage = true;
          return true;
        }
      }
      return false;
    }
  }

  public class Paddle : Ship {

    public enum State { Entry, Sleeping, Inflating, Playing };

    public bool leftPaddle;
    public State state = State.Entry;
    public int sleepCounter = 10;
    public Ball ball;
    public double angle;
    public double angle2;
    public double angleInc;
    public double angleInc2;
    public int inflatingCount = 26;
    public double entryAngle;
    public double entryAngleInc = Math.PI / 80.0;
    public int entrySteps = 40;

    public Paddle(SquariusForm squariusForm, int x, bool leftPaddle, Ball ball)
        : base(squariusForm) {
      angleInc = squariusForm.random.NextDouble() * 0.03 + 0.05;
      angleInc2 = squariusForm.random.NextDouble() * 0.015 + 0.025; 
      brush = Brushes.White;
      this.height = 128;
      this.leftPaddle = leftPaddle;
      this.x = x;
      this.ball = ball;
      if (leftPaddle) {
        y = -128;
      } else {
        y = 480;
      }
    }

    public override bool Move() {
      switch (state) {
        case State.Entry:
          if (leftPaddle) {
            y = (int)(304 * Math.Sin(entryAngle)) - 128;
          } else {
            y = 480 - (int)(304 * Math.Sin(entryAngle));
          }
          if (entrySteps > 0) {
            entrySteps--;
            if (entrySteps == 0) {
              y = 176;
              state = State.Sleeping;
            }
          }
          entryAngle += entryAngleInc;
          break;
        case State.Sleeping:
          sleepCounter--;
          if (sleepCounter == 0) {
            state = State.Inflating;
            if (leftPaddle) {
              squariusForm.ships.Add(ball);
            }
          }
          break;
        case State.Inflating:
          if (inflatingCount > 0) {
            inflatingCount--;
            if (inflatingCount == 0) {
              state = State.Playing;
            }
          }
          break;
        case State.Playing:
          angle += angleInc;
          angle2 += angleInc2;
          y = (int)(ball.y - 32 + 0.35 * (ball.x - x) * Math.Sin(angle)
              + 31 * Math.Sin(angle2));
          if (y < 32) {
            y = 64 - y;
          } else if (y > 320) {
            y = 640 - y;
          }
          break;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class PongBoss : Ship {

    public int sleepCount = 300;
    public bool movesDown;
    public bool movementComplete;

    public PongBoss(SquariusForm squariusForm, int x) 
        : base(squariusForm) {
      this.x = x;
      this.y = -480;
      this.height = 480;
      this.brush = Brushes.SandyBrown;
    }

    public override bool Move() {
      if (sleepCount > 0) {
        sleepCount--;
      } else {
        if (y < 0) {
          y += 2;
        } else if (!movementComplete) {
          movementComplete = true;
          Ball ball = new Ball(squariusForm, x + 96, 240);
          squariusForm.ships.Add(new Paddle(squariusForm, x + 32, true, ball));
          squariusForm.ships.Add(new Paddle(squariusForm, x + 576, false, ball));
        }
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Stalagtite : Ship {

    public int sleepCount = 160;
    public bool movesDown;

    public Stalagtite(SquariusForm squariusForm, int x, int y, bool movesDown) 
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.height = 128;
      this.brush = Brushes.SandyBrown;
      this.movesDown = movesDown;
    }

    public override bool Move() {
      if (sleepCount > 0) {
        sleepCount--;
      } else if (movesDown) {
        if (y + height < 480) {
          y += 4;
        }
      } else {
        if (y > 0) {
          y -= 4;
        }
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Creeper : Ship {

    public int vx;
    public int sleepCount;
    public int moveCount;
    public bool shootsDown;

    public Creeper(SquariusForm squariusForm, int x, int y, 
        bool shootsDown, bool yieldsPowerUp) : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.shootsDown = shootsDown;
      this.yieldsPowerUp = yieldsPowerUp;
      brush = yieldsPowerUp ? Brushes.Orange : Brushes.FloralWhite;
      vx = squariusForm.random.Next(2) == 0 ? 2 : -2;
      moveCount = squariusForm.random.Next(60) + 20;
    }

    public override bool Move() {
      if (sleepCount == 0) {
        if (moveCount > 0) {
          int nextX = x + vx;
          if (vx < 0) {
            if (squariusForm.IsWall(nextX, y + 16)) {
              vx = -vx;
            } else {
              x = nextX;
            }
          } else {
            if (squariusForm.IsWall(nextX + width, y + 16)) {
              vx = -vx;
            } else {
              x = nextX;
            }
          }
          moveCount--;
        } else {
          int dj = shootsDown ? 12 : -12;
          int vj = shootsDown ? 5 : -5;
          for (int i = 0, j = y + 16; i < 5; i++, j += dj) {
            squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
                x + 16, j, 0, vj));
          }
          sleepCount = squariusForm.random.Next(15) + 10;
          moveCount = squariusForm.random.Next(60) + 20;
          vx = squariusForm.random.Next(2) == 0 ? 2 : -2;
        }
      } else {
        sleepCount--;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return true;
    }
  }

  public class RightAngleShip : Ship {

    public enum State { BeforeTurn, Turning, AfterTurn, AfterShoot };

    public State state = State.BeforeTurn;
    public int targetY;

    public RightAngleShip(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      targetY = 480 - y;
      this.yieldsPowerUp = yieldsPowerUp;
      if (yieldsPowerUp) {
        brush = Brushes.Orange;
      } else {
        brush = Brushes.SeaGreen;
      }
    }

    public override bool Move() {
      switch (state) {
        case State.BeforeTurn:
          x -= 2;
          if (x < squariusForm.viewportX + 540) {
            state = State.Turning;
          }
          break;
        case State.Turning:
          x += squariusForm.scrollSpeed;
          if (Math.Abs(targetY - y) <= 4) {
            state = State.AfterTurn;
          } else if (targetY < y) {
            y -= 4;
          } else {
            y += 4;
          }
          break;
        case State.AfterTurn:
          x -= 2;
          if (x < squariusForm.viewportX + 100) {
            squariusForm.bullets.Add(new Bullet(squariusForm, this));
            state = State.AfterShoot;
          }
          break;
        case State.AfterShoot:
          x -= 2;
          break;
      }
      return x + width < squariusForm.viewportX; 
    }

    public override bool Shot() {
      return true;
    }

    public static void CreateFleet(SquariusForm squariusForm, int x, int y, 
        bool yieldsPowerUp) {
      for (int i = 0, j = 0; i < 6; i++, j += 64) {
        squariusForm.ships.Add(new RightAngleShip(
            squariusForm, x + j, y, i == 5 && yieldsPowerUp));
      }
    }
  }
}