/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_1() {
      scrollSpeed = 2;
      wallBrush = Brushes.SlateBlue;
      Text = "SQUARIUS\u00B2 - Stage 2";
    }

    public void ProcessCell_1(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'B':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'C':
          RightAngleShip.CreateFleet(this, x, y, false);
          break;
        case 'D':
          RightAngleShip.CreateFleet(this, x, y, true);
          break;
        case 'E':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'F':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'G':
          ships.Add(new Bubble(this, x, y, 2));
          break;
        case 'H':
          ships.Add(new Bubble(this, x, y, 3));
          break;
        case 'I':
          ships.Add(new Bubble(this, x, y, 4));
          break;
        case 'J':
          ships.Add(new Bubble(this, x, y, 5));
          break;
        case 'Y':
          ships.Add(new BlobBoss(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class BlobBoss : Ship {

    public enum State { Paused, Bouncing, Stretching, Shooting };

    public int paused = 200;
    public Brush defaultBrush;
    public double Y;
    public double DY;
    public State state = State.Paused;
    public double angle;
    public double scale;
    public int X;
    public int flashing;
    public int shotCount;
    public int shootPause;
    public int blobCount;
    public double DX;
    public double XX;

    public BlobBoss(SquariusForm squariusForm, int x, int y) 
        : base(squariusForm) {
      XX = X = this.x = x;
      Y = this.y = -256;
      defaultBrush = brush = Brushes.GreenYellow;
      width = height = 256;
    }

    public override bool Move() {
      if (state == State.Paused) {
        paused--;
        if (paused == 0) {
          state = State.Bouncing;
        }
        return false;
      } else if (state == State.Bouncing) {
        DY += 0.2;
        XX += DX;
        x = (int)XX;
        Y += DY;
        y = (int)Y;
        if (y >= 192 && DY > 0) {
          y = 192;
          state = State.Stretching;
          angle = 0;
          scale = 64;
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
        }
      } else if (state == State.Stretching) {
        angle += 0.2;
        scale *= 0.96;
        int stretch = (int)(scale * Math.Sin(angle));
        y = 192 + stretch;
        x = (int)(XX - (stretch >> 1));
        width = 256 + stretch;
        height = 256 - stretch;
        if (scale < 0.9) {
          state = State.Shooting;
        }
      } else if (state == State.Shooting) {
        if (shootPause <= 0) {
          squariusForm.ships.Add(new Bubble(
            squariusForm, 
            x + squariusForm.random.Next(128), 
            y + squariusForm.random.Next(128), 2));
          shootPause = 10;
          blobCount++;
          if (blobCount == 10) {
            blobCount = 0;
            state = State.Bouncing;
            DY = -8;
            int targetX = squariusForm.random.Next(
                squariusForm.viewportX + 128, X + 64);
            DX = (targetX - x) / 79.0;
          }
        } else {
          shootPause--;
        }
      }
      if (flashing > 0) {
        flashing--;
        brush = Brushes.White;
      } else {
        brush = defaultBrush;
      }
      return false;      
    }

    public override bool Shot() {
      if (flashing == 0 && state == State.Stretching) {
        flashing = 2;
        shotCount++;
        if (shotCount > 105) {
          squariusForm.requestAdvanceStage = true;
          return true;
        } else if (shotCount > 80) {
          brush = defaultBrush = Brushes.Magenta;
        }
      }
      return false;
    }
  }

  public class Bubble : Ship {

    public double angleX;
    public double angleY;
    public double angleDX;
    public double angleDY;
    public int WIDTH;
    public int HEIGHT;
    public int vx;
    public int vy;
    public int size;
    public int shots = 3;
    public int flashing;
    public Brush defaultBrush;

    public Bubble(SquariusForm squariusForm, int x, int y, int size) 
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.vx = -2;
      this.vy = squariusForm.random.Next(-3, 3);
      this.size = size;
      angleX = squariusForm.random.NextDouble() * Math.PI * 2;
      angleY = squariusForm.random.NextDouble() * Math.PI * 2;
      angleDX = 0.01 + 0.1 * squariusForm.random.NextDouble();
      angleDY = 0.01 + 0.1 * squariusForm.random.NextDouble();
      HEIGHT = WIDTH = size << 5;      
      ComputeWidthAndHeight();

      switch (size) {
        case 2:
          brush = Brushes.GreenYellow;
          break;
        case 3:
          brush = Brushes.Lime;
          break;
        case 4:
          brush = Brushes.Chartreuse;
          break;
        default:
          brush = Brushes.MediumSeaGreen;
          break;
      }
      defaultBrush = brush;
    }

    private void ComputeWidthAndHeight() {
      width = WIDTH + (int)(16 * Math.Sin(angleX));
      height = HEIGHT + (int)(16 * Math.Sin(angleY));
    }

    public override bool Move() {
      angleX += angleDX;
      angleY += angleDY;
      ComputeWidthAndHeight();
      if (y <= 32) {
        vy = Math.Abs(vy);
      } else if (y + height >= 448) {
        vy = -Math.Abs(vy);
      }
      x += vx;
      y += vy;
      if (flashing > 0) {
        flashing--;
        brush = Brushes.White;
      } else {
        brush = defaultBrush;
      }
      return x + width < squariusForm.viewportX;      
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 1;
        shots--;
      } else {
        return false;
      }
      if (shots < 0) {
        if (size > 2) {          
          squariusForm.ships.Add(new Bubble(squariusForm, x, y, size - 1));
          squariusForm.ships.Add(new Bubble(squariusForm, 
              x + (width >> 1), y + (height >> 1), size - 1));
        }
        return true;
      }
      return false;
    }
  }
}