/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_2() {
      scrollSpeed = 2;
      wallBrush = Brushes.Peru;
      Text = "SQUARIUS\u00B2 - Stage 3";
    }

    public void ProcessCell_2(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'B':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'C':
          ships.Add(new Creeper(this, x, y, true, false));
          break;
        case 'D':
          ships.Add(new Creeper(this, x, y, false, false));
          break;
        case 'E':
          WaveShip.CreateWave(this, x, y, 64, false);
          break;
        case 'F':
          WaveShip.CreateWave(this, x, y, 64, true);
          break;
        case 'G':
          WaveShip.CreateHelix(this, x, y, 64, false);
          break;
        case 'H':
          WaveShip.CreateHelix(this, x, y, 64, true);
          break;
        case 'I':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'J':
          ships.Add(new Lava(this, x, y));
          break;
        case 'K':
          ships.Add(new WallShip(this, x, y));
          break;
        case 'L':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'M':
          ships.Add(new WallShip(this, x, y, 16));
          break;
        case 'N':
          ships.Add(new HorizontalShipGenerator(this, x, y));
          break;
        case 'O':
          ships.Add(new WallLazer(this, x, y, y + 128));
          break;
        case 'P':
          ships.Add(new WallLazer(this, x, y, y - 128));
          break;
        case 'Q':
          ships.Add(new CaveBoss(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class Lava : Ship {

    public int rockCount;

    public Lava(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      width = 96;
      this.x = x;
      this.y = y;
      brush = Brushes.White;
    }

    public override bool Move() {
      if (rockCount == 0) {
        rockCount = 30 + squariusForm.random.Next(30);
        squariusForm.ships.Add(new Rock(squariusForm, x + 32, y));
      } else {
        rockCount--;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Rock : Ship {

    public int dx;
    public double dy;
    public double Y;

    public Rock(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      dx = squariusForm.random.Next(-2, 3);
      dy = squariusForm.random.Next(-10,-6);
      this.x = x;
      Y = this.y = y;
      brush = Brushes.White;
    }

    public override bool Move() {
      x += dx;
      Y += dy;
      dy += 0.2;
      y = (int)Y;
      return x + width < squariusForm.viewportX || y > 480;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class WaveShip : Ship {

    public int scale;
    public int Y;
    public int shootCount;

    public WaveShip(SquariusForm squariusForm, int x, int y, 
        int scale, bool yieldsPowerUp) : base(squariusForm) {
      brush = yieldsPowerUp ? Brushes.Orange : Brushes.DarkKhaki;
      this.x = x;
      Y = y;
      this.y = Y + (int)(scale * Math.Sin((squariusForm.viewportX - x) / 64.0));
      this.scale = scale;
      this.yieldsPowerUp = yieldsPowerUp;
    }

    public override bool Move() {
      x -= 2;
      y = Y + (int)(scale * Math.Sin((squariusForm.viewportX - x) / 64.0));
      if (shootCount == 0) {
        squariusForm.bullets.Add(new Bullet(squariusForm, this));
        shootCount = 80;
      } else {
        shootCount--;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return true;
    }

    public static void CreateWave(SquariusForm squariusForm, int x, int y,
        int scale, bool yieldsPowerUp) {
      for (int i = 0; i < 10; i++) {
        squariusForm.ships.Add(new WaveShip(squariusForm, x + i * 64, y, scale,
            i == 9 && yieldsPowerUp));
      }
    }

    public static void CreateHelix(SquariusForm squariusForm, int x, int y,
        int scale, bool yieldsPowerUp) {
      for (int i = 0; i < 10; i++) {
        squariusForm.ships.Add(new WaveShip(squariusForm, x + i * 64, y, scale,
            i == 9 && yieldsPowerUp));
        squariusForm.ships.Add(new WaveShip(squariusForm, x + i * 64, y, -scale,
            false));
      }
    }
  }

  public class HorizontalShipGenerator : Ship {

    public int sleepCount;

    public HorizontalShipGenerator(SquariusForm squariusForm, int x, int y) 
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.brush = Brushes.Magenta;
      sleepCount = squariusForm.random.Next(40, 120);
    }

    public override bool Move() {
      if (sleepCount == 0) {
        squariusForm.ships.Add(new HorizontalShip(squariusForm, x, y, false, true));
        sleepCount = squariusForm.random.Next(120, 280);
      } else {
        sleepCount--;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class CaveBoss : Ship {
        
    public int hits = 30;
    public int flashing;
    public Brush defaultBrush;

    public CaveBoss(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      defaultBrush = brush = Brushes.SpringGreen;
      this.x = x;
      this.y = y;
    }

    public override bool Move() {
      if (flashing > 0) {
        flashing--;
        brush = Brushes.White;
      } else {
        brush = defaultBrush;
      }

      return false;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        hits--;
        if (hits == 25) {
          defaultBrush = Brushes.Magenta;
        } else if (hits == 0) {
          squariusForm.requestAdvanceStage = true;
          return true;
        }
      } 
      return false;
    }
  }

  public class WallLazer : Ship {

    public enum State { Sleeping, Move1, Move2 }

    public int sleepCount = 280;
    public State state = State.Sleeping;
    public int dy;
    public int targetY;
    public int shootDelay = 5;
    public int moveCount;

    public WallLazer(SquariusForm squariusForm, int x, int y, int targetY) 
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.targetY = targetY;
      brush = Brushes.MediumOrchid;
      if (targetY > y) {
        dy = 1;
      } else {
        dy = -1;
      }
    }

    public override bool Move() {
      switch (state) {
        case State.Sleeping:
          if (sleepCount == 0) {
            state = State.Move1;
          } else {
            sleepCount--;
          }
          break;
        case State.Move1:
          y += dy;
          moveCount++;
          if (y == targetY) {
            state = State.Move2;
          }
          if (shootDelay == 0) {
            shootDelay = 3;
            squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
                x + 16, y + 16, -5, 0));
          } else {
            shootDelay--;
          }
          break;
        case State.Move2:
          if (moveCount > 0) {
            y -= dy;
            moveCount--;
          } else {
            state = State.Sleeping;
            sleepCount = 320;
          }
          if (shootDelay == 0) {
            shootDelay = 3;
            squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
                x + 16, y + 16, -5, 0));
          } else {
            shootDelay--;
          }
          break;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class WallShip : Ship {

    public int hits;
    public int flashing;
    public Brush defaultBrush;

    public WallShip(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      defaultBrush = brush = squariusForm.wallBrush;
      this.x = x;
      this.y = y;
      this.hits = 0;
    }

    public WallShip(SquariusForm squariusForm, int x, int y, int hits)
          : base(squariusForm) {
      defaultBrush = brush = squariusForm.wallBrush;
      this.x = x;
      this.y = y;
      this.hits = hits;
    }

    public override bool Move() {
      if (flashing > 0) {
        brush = Brushes.White;
        flashing--;
      } else {
        brush = defaultBrush;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        if (hits <= 0) {
          return true;
        } else {
          hits--;
        }
      }
      return false;
    }
  }
}