/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_3() {
      scrollSpeed = 2;
      wallBrush = Brushes.Silver;
      Text = "SQUARIUS\u00B2 - Stage 4";
    }

    public void ProcessCell_3(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          WaveShip.CreateWave(this, x, y, 64, false);
          break;
        case 'B':
          WaveShip.CreateWave(this, x, y, 64, true);
          break;
        case 'C':
          WaveShip.CreateHelix(this, x, y, 64, false);
          break;
        case 'D':
          WaveShip.CreateHelix(this, x, y, 64, true);
          break;
        case 'E':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'F':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'G':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'H':
          ships.Add(new Brick(this, x, y, 192, 192, 160, y + 160));
          break;
        case 'I':
          ships.Add(new LargeWallGun(this, x, y, false));
          break;
        case 'J':
          ships.Add(new Brick(this, x, y, 192, 192, 160, 256));
          break;
        case 'K':
          ships.Add(new LargeWallGun(this, x, y, true));
          break;
        case 'L':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'Y':
          ships.Add(new BrickBoss(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class BrickBoss : Ship {

    public Brush defaultBrush;
    public int pauseDelay = 280;
    public double angle;
    public int rocking;
    public bool volunerable;
    public int flashing;
    public int shotCount;

    public BrickBoss(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      width = height = 128;
      defaultBrush = brush = Brushes.RosyBrown;
    }

    public override bool Move() {
      if (pauseDelay > 0) {
        pauseDelay--;
        if (pauseDelay == 0) {
          volunerable = true;
          if (squariusForm.random.Next(2) == 0) {
            angle = 2.0 * squariusForm.random.Next(1, 3) * Math.PI;
          } else {
            rocking = 200;
          }
        }
      } else if (angle > 0) {
        squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
            x + 64, y + 64, 5 * Math.Cos(angle), 5 * Math.Sin(angle)));
        angle -= 0.06;
        if (angle <= 0) {
          pauseDelay = 60;
        }
      } else if (rocking > 0) {
        if ((rocking & 15) == 0) {
          squariusForm.ships.Add(new Rock(squariusForm, x + 48, y - 32));
        }
        rocking--;
        if (rocking == 0) {
          pauseDelay = 60;
        }
      }
      if (flashing > 0) {
        flashing--;
        brush = Brushes.White;
      } else {
        brush = defaultBrush;
      }
      return false;
    }

    public override bool Shot() {
      if (!volunerable) {
        return false;
      }
      if (flashing == 0) {
        flashing = 2;
        shotCount++;
        if (shotCount > 120) {
          squariusForm.requestAdvanceStage = true;
          return true;
        } else if (shotCount > 90) {
          brush = defaultBrush = Brushes.Magenta;
        }
      }
      return false;
    }
  }

  public class Brick : Ship {

    public int dropDelay;
    public int targetY;
    public double Y;
    public double DY;

    public Brick(SquariusForm squariusForm, int x, int y, int width, int height, 
        int dropDelay, int targetY) : base(squariusForm) {
      this.x = x;
      Y = this.y = y;
      DY = 1;
      this.width = width;
      this.height = height;
      this.targetY = targetY;
      this.brush = squariusForm.wallBrush;
      this.dropDelay = dropDelay;
    }

    public override bool Move() {
      if (dropDelay == 0) {
        if (y != targetY) {
          Y += DY;
          DY += 0.2;
          y = (int)Y;
          if (y >= targetY) {
            y = targetY;
          }
        }
      } else {
        dropDelay--;
      }
      return x + width < squariusForm.viewportX; 
    }

    public override bool Shot() {
      return false;
    }
  }

  public class LargeWallGun : Ship {

    public int shootCounter = 10;
    public int flashing;
    public int hits = 12;
    public Brush defaultBrush;

    public LargeWallGun(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp)
        : base(squariusForm) {
      defaultBrush = brush = yieldsPowerUp ? Brushes.Orange : Brushes.RosyBrown;
      this.x = x;
      this.y = y;
      this.yieldsPowerUp = yieldsPowerUp;
      this.height = this.width = 96;
    }

    public override bool Move() {
      if (shootCounter == 0) {
        shootCounter = 20;
        squariusForm.bullets.Add(new Bullet(squariusForm, this));
      } else {
        shootCounter--;
      }
      if (flashing > 0) {
        brush = (flashing & 1) == 0 ? Brushes.White : defaultBrush;
        flashing--;
      } else {
        brush = defaultBrush;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        hits--;
        if (hits == 0) {
          return true;
        }
      }
      return false;
    }
  }
}