/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_4() {
      scrollSpeed = 2;
      wallBrush = Brushes.DarkOrange;
      Text = "SQUARIUS\u00B2 - Stage 5";
    }

    public void ProcessCell_4(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new Missile(this, x, y, false));
          break;
        case 'B':
          ships.Add(new Missile(this, x, y, true));
          break;
        case 'C':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'D':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'E':
          CircleShip.CreateRing(this, x, y, 128, 8, false);
          break;
        case 'F':
          CircleShip.CreateRing(this, x, y, 128, 8, true);
          break;
        case 'G':
          CircleShip.CreateSpiral(this, x, y, 192, 16, true);
          break;
        case 'H':
          ships.Add(new Boulder(this, x, y, 2, -10, 0));
          break;
        case 'I':
          ships.Add(new Creeper(this, x, y, false, false));
          break;
        case 'J':
          ships.Add(new Creeper(this, x, y, true, false));
          break;
        case 'K':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'L':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'M':
          ships.Add(new Balloon(this, x, y, 4));
          break;
        case 'N':
          ships.Add(new Balloon(this, x, y, 6));
          break;
        case 'O':
          ships.Add(new Balloon(this, x, y, 8));
          break;
        case 'P':
          ships.Add(new SeekerShip(this, x - 672, y, false));
          break;
        case 'Q':
          ships.Add(new SeekerShip(this, x - 672, y, true));
          break;
        case 'R':
          ships.Add(new SeekerShip(this, x, y, false));
          break;
        case 'Y':
          ships.Add(new Snake(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class Snake : Ship {

    public Tail tail;
    public double angle = Math.PI;
    public int minX;
    public int maxX;
    public int centerX;
    public double drift;
    public int flashing;
    public Brush defaultBrush;
    public int shootDelay = 100;
    public int hits;
    public int explodeDelay;

    public Snake(SquariusForm squariusForm, int x, int y) : base(squariusForm) {      
      this.x = x - 736;
      this.y = y;
      minX = this.x;
      maxX = x - 32;
      centerX = (maxX + minX) >> 1;
      this.width = this.height = 64;
      defaultBrush = this.brush = Brushes.MediumTurquoise;
      tail = new Tail(squariusForm, this.x, this.y, width, height, 8, 32, brush, this);
    }

    public override bool Move() {
      if (hits < 300) {
        angle += 0.01;
        drift += 0.002;
        x = (int)(centerX + 320 * Math.Cos(angle));
        y = (int)(208 + 132 * Math.Sin(angle * 2 + drift)
            + 44 * Math.Sin(angle * 5 - drift));
        if (shootDelay == 0) {
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
          shootDelay = 100;
        } else {
          shootDelay--;
        }
        if (flashing > 0) {
          flashing--;
          if (flashing == 0) {
            tail.Brush = defaultBrush;
            brush = defaultBrush;
          }
        }
      }
      tail.Follow(x, y);
      if (tail.exploded && squariusForm.playerAlive) {
        squariusForm.requestAdvanceStage = true;        
      }
      return tail.exploded;
    }

    public override bool Shot() {
      if (flashing == 0 && hits < 300) {
        flashing = 2;
        tail.Brush = brush = Brushes.White;
        hits++;
        if (hits == 225) {
          defaultBrush = Brushes.Magenta;
        } else if (hits == 300) {
          tail.Explode();
        }        
      }
      return tail.exploded;
    }
  }

  public class TailShip : Ship {

    public bool explode;
    public Ship head;

    public TailShip(SquariusForm squariusForm, Ship head) : base(squariusForm) {
      this.head = head;
    }

    public override bool Move() {
      return explode;
    }

    public override bool Shot() {
      head.Shot();
      return explode;
    }
  }

  public class Tail {

    public SquariusForm squariusForm;
    public TailShip[] ships;
    public Point2D[] points;
    public int[] indexes;
    public bool exploding;
    public int explodeIndex;
    public bool exploded;

    public Tail(SquariusForm squariusForm, int x, int y, int width, int height,
        int spacing, int count, Brush brush, Ship head) {
      this.squariusForm = squariusForm;
      points = new Point2D[count * spacing];
      for (int i = 0; i < points.Length; i++) {
        points[i] = new Point2D(x, y);
      }
      ships = new TailShip[count];
      indexes = new int[count];
      for (int i = 0, j = 0; i < ships.Length; i++, j += spacing) {
        TailShip ship = new TailShip(squariusForm, head);
        ship.x = x;
        ship.y = y;
        ship.width = width;
        ship.height = height;
        ship.brush = brush;
        ships[i] = ship;
        if (i > 0) {
          squariusForm.ships.Add(ship);
        }
        indexes[i] = j;
      }      
    }

    public Brush Brush {
      set {
        foreach (TailShip ship in ships) {
          ship.brush = value;
        }
      }
      get {
        return ships[0].brush;
      }
    }

    public void Follow(int x, int y) {
      if (exploding) {        
        if (explodeIndex >= 0) {
          TailShip ship = ships[explodeIndex--];
          ship.explode = true;
          squariusForm.explosions.Add(new Explosion(squariusForm, ship));
          if (explodeIndex < 0) {
            exploded = true;
          }
        }
      } else {
        for (int i = 0; i < indexes.Length; i++) {
          if (indexes[i] == 0) {
            indexes[i] = points.Length - 1;
          } else {
            indexes[i]--;
          }
          TailShip s = ships[i];
          Point2D p = points[indexes[i]];
          s.x = p.x;
          s.y = p.y;
        }
        Point2D point = points[indexes[0]];
        point.x = x;
        point.y = y;
        TailShip ship = ships[0];
        ship.x = x;
        ship.y = y;
      }
    }

    public void Explode() {
      explodeIndex = ships.Length - 1;
      exploding = true;
    }
  }

  public class SeekerShip : Ship {

    public double ax;
    public double ay;
    public double vx;
    public double vy;

    public SeekerShip(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.brush = yieldsPowerUp ? Brushes.Orange : Brushes.Aquamarine;
      this.yieldsPowerUp = yieldsPowerUp;
    }

    public override bool Move() {
      Ship playerShip = squariusForm.playerShip;
      double ax = playerShip.x - x;
      double ay = playerShip.y - y;
      double mag = Math.Sqrt(ax * ax + ay * ay);
      if (mag > 0) {
        double scale = 0.3 / mag;
        ax *= scale;
        ay *= scale;
        vx += ax;
        vy += ay;
        vx *= 0.98;
        vy *= 0.98;
        x = (int)(x + vx);
        y = (int)(y + vy);
      } else {
        return true;
      }
      return false;
    }

    public override bool Shot() {
      return true;
    }
  }

  public class CircleShip : Ship {

    public double radius;
    public double angle;
    public int shootDelay = 80;
    public int X;
    public int Y;

    public CircleShip(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp,
        double radius, double angle) : base(squariusForm) {
      this.X = x;
      this.Y = y;
      this.x = (int)(X + radius * Math.Cos(angle));
      this.y = (int)(Y + radius * Math.Sin(angle));
      this.radius = radius;
      this.angle = angle;
      this.yieldsPowerUp = yieldsPowerUp;
      if (yieldsPowerUp) {
        brush = Brushes.Orange;
      } else {
        brush = Brushes.SeaGreen;
      }
    }

    public override bool Move() {
      if (shootDelay > 0) {
        shootDelay--;
      } else {
        shootDelay = 120;
        squariusForm.bullets.Add(new Bullet(squariusForm, this));
      }
      angle += 0.025;
      this.x = (int)(X + radius * Math.Cos(angle));
      this.y = (int)(Y + radius * Math.Sin(angle));
      X -= 2;

      return x + width + radius < squariusForm.viewportX;
    }

    public override bool Shot() {
      return true;
    }

    public static void CreateRing(SquariusForm squariusForm, int x, int y,
        double radius, int count, bool yieldsPowerUp) {
      double angleInc = 2.0 * Math.PI / count;
      double angle = 0;
      for (int i = 0; i < count; i++, angle += angleInc) {
        CircleShip ship = new CircleShip(squariusForm, x, y,
            yieldsPowerUp && (i == 0), radius, angle);
        ship.X += (int)radius;
        squariusForm.ships.Add(ship);
      }
    }

    public static void CreateSpiral(SquariusForm squariusForm, int x, int y,
        double radius, int count, bool yieldsPowerUp) {
      double angleInc = 2.0 * Math.PI / count;
      double radiusInc = radius / count;
      double R = radius;
      double angle = 0;
      for (int i = 0; i < count; i++, angle += angleInc, radius -= radiusInc) {
        CircleShip ship = new CircleShip(squariusForm, x, y,
            yieldsPowerUp && (i + 1 == count), radius, angle);
        ship.X += (int)R;
        squariusForm.ships.Add(ship);
      }
    }
  }

  public class Balloon : Ship {

    public int size;

    public Balloon(SquariusForm squariusForm, int x, int y, int size) 
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      size <<= 5;
      this.size = size;
      this.brush = Brushes.SlateBlue;
      this.width = this.height = size;
    }

    public override bool Move() {
      if (width < size) {
        x -= 2;
        y -= 2;
        width += 4;
        height += 4;
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      if (width > 32) {
        x += 2;
        y += 2;
        width -= 4;
        height -= 4;
      }
      return false;
    }
  }

  public class Boulder : Ship {

    public int vx;
    public int vy;
    public int size;
    public int delay = 2;

    public Boulder(SquariusForm squariusForm, int x, int y, int size,
        int vx, int vy) : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.vx = vx;
      this.vy = vy;
      this.size = size;
      width = height = size << 5;
      brush = squariusForm.wallBrush;
    }

    public override bool Move() {
      if (delay > 0) {
        delay--;
      }
      x += vx;
      y += vy;
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      if (delay > 0) {
        return false;
      }
      if (size > 1) {
        size--;
        int h = height / 3;
        vx++;
        vy += squariusForm.random.Next(-1, 1);
        squariusForm.ships.Add(new Boulder(squariusForm,
            x + h, y, size, vx, vy - 2));
        squariusForm.ships.Add(new Boulder(squariusForm, 
            x, y + h, size, vx, vy));
        squariusForm.ships.Add(new Boulder(squariusForm,
            x + h, y + (h << 1), size, vx, vy + 2));
      }
      return true;
    }
  }

  public class Missile : Ship {

    public int delay = 10;

    public Missile(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.width = 128;
      this.yieldsPowerUp = yieldsPowerUp;
      if (yieldsPowerUp) {
        brush = Brushes.Orange;
      } else {
        brush = Brushes.CornflowerBlue;
      }
    }

    public override bool Move() {
      if (delay > 0) {
        delay--;
        return false;
      }
      x -= 16;
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return true;
    }
  }
}