/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_5() {
      scrollSpeed = 2;
      wallBrush = Brushes.ForestGreen;
      Text = "SQUARIUS\u00B2 - Stage 6";
    }

    public void ProcessCell_5(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'B':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'C':
          CircleShip.CreateRing(this, x, y, 128, 8, false);
          break;
        case 'D':
          CircleShip.CreateRing(this, x, y, 128, 8, true);
          break;
        case 'E':
          ships.Add(new TailSeekerShip(this, x, y, false));
          break;
        case 'F':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'G':
          ships.Add(new WallGun(this, x, y, true));
          break;
        case 'H':
          ships.Add(new Drip(this, x, y));
          break;
        case 'I':
          ships.Add(new Vine(this, x, y));
          break;
        case 'J':
          ships.Add(new SliverVine(this, x, y));
          break;
        case 'Y':
          ships.Add(new PlantBoss(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class PlantBoss : Ship {

    public enum State { Paused1, Paused2, Sucking, Seekers, Vining, Exploding, Halt };

    public int flashing;
    public int hits;
    public Brush defaultBrush;
    public int pauseDelay = 200;
    public int seekerDelay;
    public State state = State.Paused1;
    public bool seekerUp;
    public int seekerCount = 10;
    public Ribbon ribbon;
    public double radius;
    public double angle;
    public int suckingCount = 400;
    public int shootDelay = 160;
    public double radiusAngle;
    public bool fullyExtended;

    public PlantBoss(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      this.x = x;
      this.y = y;
      defaultBrush = brush = squariusForm.wallBrush;
      width = height = 128;
      ribbon = new Ribbon(squariusForm, 64, 64, 20, brush, this, true, 0,
          2.0 * Math.PI, 0);
      ribbon.Update(x + 32, y + 32, x + 32, y + 32);
    }

    public override bool Move() {
      switch (state) {
        case State.Paused1:
          if (--pauseDelay <= 0) {
            state = State.Seekers;
          }
          break;
        case State.Paused2:
          if (--pauseDelay <= 0) {
            state = State.Vining;
          }
          break;
        case State.Seekers:
          if (seekerDelay > 0) {
            seekerDelay--;
          } else {
            seekerDelay = 40;
            seekerUp = !seekerUp;
            squariusForm.ships.Add(
                new SeekerShip(squariusForm, x + 160,
                seekerUp ? 0 : 448, false));
            if (--seekerCount == 0) {
              state = State.Sucking;
            }
          }
          break;
        case State.Sucking:
          if (suckingCount > 0) {
            suckingCount--;
            if ((suckingCount & 1) == 0) {
              if (suckingCount == 50) {
                squariusForm.bullets.Add(new Bullet(squariusForm, this));
              }
              brush = defaultBrush;
              Ship playerShip = squariusForm.playerShip;
              playerShip.x += 16;
              if (playerShip.y > 240) {
                playerShip.y -= 8;
              } else {
                playerShip.y += 8;
              }
            } else {
              brush = Brushes.LawnGreen;
            }
          } else {
            state = State.Paused2;
            pauseDelay = 30;
            brush = defaultBrush;            
          }
          break;
        case State.Vining:
          if (flashing > 0) {
            flashing--;
            if (flashing == 0) {
              ribbon.Brush = defaultBrush;
              brush = defaultBrush;
            }
          }
          if (shootDelay > 0) {
            shootDelay--;
          } else {
            shootDelay = 160;
            squariusForm.bullets.Add(new Bullet(squariusForm, this));
            seekerUp = !seekerUp;
            squariusForm.ships.Add(
                new SeekerShip(squariusForm, x + 160,
                seekerUp ? 0 : 448, false));
          }
          if (fullyExtended) {
            radius = 350 + 50 * Math.Sin(radiusAngle);
            radiusAngle += 0.15;
          } else {
            radius += 10.0;
            if (radius >= 350) {
              fullyExtended = true;
            }
          }
          angle += 0.05;
          double ang = Math.PI + 0.5 * Math.Sin(angle) 
              + 0.4 * Math.Cos(0.4 * angle + 0.2);
          ribbon.waveScale = 64 * Math.Sin(2.75 * angle);
          ribbon.fullAngle = 2.0 * Math.PI + ang;
          ribbon.Update(x + 32, y + 32,
              (int)(32 + x + radius * Math.Cos(ang)),
              (int)(32 + y + radius * Math.Sin(ang)));
          break;
        case State.Exploding:
          if (ribbon.exploded && squariusForm.playerAlive) {
            squariusForm.requestAdvanceStage = true;
            state = State.Halt;
            squariusForm.explosions.Add(new Explosion(squariusForm, this));
            return true;
          } else {
            double ang1 = Math.PI + 0.5 * Math.Sin(angle)
                          + 0.4 * Math.Cos(0.4 * angle + 0.2);
            ribbon.waveScale = 64 * Math.Sin(2.75 * angle);
            ribbon.fullAngle = 2.0 * Math.PI + ang1;
            ribbon.Update(x + 32, y + 32,
                (int)(32 + x + radius * Math.Cos(ang1)),
                (int)(32 + y + radius * Math.Sin(ang1)));
          }
          break;
      }
      return false;
    }

    public override bool Shot() {
      if (state == State.Vining) {
        if (flashing == 0) {
          flashing = 2;
          ribbon.Brush = Brushes.White;
          brush = Brushes.White;
          hits++;
          if (hits == 200) {
            defaultBrush = Brushes.Magenta;
          } else if (hits == 250) {
            state = State.Exploding;
            ribbon.Brush = defaultBrush;
            brush = defaultBrush;
            ribbon.Explode();
          }
        }
      }
      return false;
    }
  }

  public class Vine : Ship {

    public int X;
    public int Y;
    public double angle;
    public int flashing;
    public int hits;
    public bool exploding;
    public Line upperLine;
    public Line lowerLine;
    public Brush defaultBrush;
    public int shootDelay = 50;

    public Vine(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      this.X = this.x = x;
      this.Y = this.y = y;
      width = height = 96;
      defaultBrush = brush = squariusForm.wallBrush;
      upperLine = new Line(squariusForm, 32, 32, 16, brush, this, false);
      upperLine.Update(x + 32, y + 32, X + 32, 32);
      lowerLine = new Line(squariusForm, 32, 32, 16, brush, this, false);
      lowerLine.Update(x + 32, y + 32, X + 32, 448);
      Move();
    }

    public override bool Move() {
      if (!exploding) {
        if (shootDelay > 0) {
          shootDelay--;
        } else {
          shootDelay = 50;
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
        }
        if (flashing > 0) {
          flashing--;
          if (flashing == 0) {
            upperLine.Brush = defaultBrush;
            lowerLine.Brush = defaultBrush;
            brush = defaultBrush;
          }
        }
        x = (int)(X - 96 * Math.Abs(Math.Sin(angle)));
        y = (int)(Y + 96 * Math.Sin(4.0 * angle));        
        angle += 0.025;
      }
      upperLine.Update(x + 32, y + 32, X + 32, 32);
      lowerLine.Update(x + 32, y + 32, X + 32, 448);
      if (upperLine.exploded && lowerLine.exploded) {
        squariusForm.explosions.Add(new Explosion(squariusForm, this));
        return true;
      }
      return X + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      if (flashing == 0 && !exploding) {
        flashing = 2;
        hits++;
        upperLine.Brush = Brushes.White;
        lowerLine.Brush = Brushes.White;
        brush = Brushes.White;
        if (hits == 30) {
          exploding = true;
          upperLine.Explode();
          lowerLine.Explode();
        }
      }
      return upperLine.exploded && lowerLine.exploded;
    }
  }

  public class SliverVine : Ship {

    public Ribbon ribbon;
    public double seekX;
    public double seekY;
    public bool shot;
    public double angle;

    public SliverVine(SquariusForm squariusForm, int x, int y) 
        : base(squariusForm) {
      seekX = this.x = x;
      seekY = this.y = y;
      this.brush = squariusForm.wallBrush;
      ribbon = new Ribbon(squariusForm, 32, 32, 16, brush, this, true,
          0, 2.0 * Math.PI, 32);
      ribbon.Update(x, y, (int)seekX, (int)seekY);
    }

    public override bool Move() {
      Ship playerShip = squariusForm.playerShip;
      double vx = playerShip.x - seekX;
      double vy = playerShip.y - seekY;
      double scale = 1.5 / Math.Sqrt(vx * vx + vy * vy);
      if (shot) {
        scale = -scale;
        shot = false;
      }
      vx *= scale;
      vy *= scale;
      seekX += vx;
      seekY += vy;
      angle += 0.02;
      if (x + width < squariusForm.viewportX) {
        ribbon.Explode();
      }
      ribbon.waveScale = 32 * Math.Sin(angle);
      ribbon.Update(x, y, (int)seekX, (int)seekY);
      return ribbon.exploded;
    }

    public override bool Shot() {
      shot = true;
      return false;
    }
  }

  public class Ribbon {

    public SquariusForm squariusForm;
    public TailShip[] ships;
    public bool exploding;
    public bool exploded;
    public int explodeIndex;
    public double angleOffset;
    public double fullAngle;
    public double waveScale;
    public double scale;

    public Ribbon(SquariusForm squariusForm, int width, int height,
        int count, Brush brush, Ship parent, bool includeParent,
        double angleOffset, double fullAngle, double waveScale) {
      this.squariusForm = squariusForm;
      this.angleOffset = angleOffset;
      this.fullAngle = fullAngle;
      this.waveScale = waveScale;
      scale = 1.0 / count;
      ships = new TailShip[count];
      for (int i = 0; i < count; i++) {
        TailShip ship = new TailShip(squariusForm, parent);
        ship.brush = brush;
        ship.width = width;
        ship.height = height;
        if (i > 0 || includeParent) {
          squariusForm.ships.Add(ship);
        }
        ships[i] = ship;
      }
    }

    public Brush Brush {
      get {
        return ships[0].brush;
      }
      set {
        foreach (Ship ship in ships) {
          ship.brush = value;
        }
      }
    }

    public void Update(int x1, int y1, int x2, int y2) {
      if (exploding) {
        if (!exploded) {
          ships[explodeIndex].explode = true;
          squariusForm.explosions.Add(
              new Explosion(squariusForm, ships[explodeIndex]));
          if (explodeIndex == 0) {
            exploded = true;
          } else {
            explodeIndex--;
          }
        }
      } else {
        double vx = x2 - x1;
        double vy = y2 - y1;
        double dx = scale * vx;
        double dy = scale * vy;
        double vMag = Math.Sqrt(vx * vx + vy * vy);
        if (vMag > 0) {
          double vScale = 1.0 / vMag;
          vx *= vScale;
          vy *= vScale;
          double ux = -vy;
          double uy = vx;
          double x = x1;
          double y = y1;
          double angle = angleOffset;
          double angleInc = fullAngle / ships.Length;
          for (int i = 0; i < ships.Length; i++, angle += angleInc) {
            x += dx;
            y += dy;
            double sin = waveScale * Math.Sin(angle);
            ships[i].x = (int)(x + ux * sin);
            ships[i].y = (int)(y + uy * sin);
          }
        }
      }
    }

    public void Explode() {
      if (exploding) {
        return;
      }
      exploding = true;
      explodeIndex = ships.Length - 1;
    }
  }

  public class Line {

    public SquariusForm squariusForm;
    public TailShip[] ships;
    public double scale;
    public bool exploding;
    public bool exploded;
    public int explodeIndex;

    public Line(SquariusForm squariusForm, int width, int height,
        int count, Brush brush, Ship parent, bool includeParent) {
      this.squariusForm = squariusForm;
      scale = 1.0 / count;
      ships = new TailShip[count];
      for (int i = 0; i < count; i++) {
        TailShip ship = new TailShip(squariusForm, parent);
        ship.brush = brush;
        ship.width = width;
        ship.height = height;
        if (i > 0 || includeParent) {
          squariusForm.ships.Add(ship);
        }
        ships[i] = ship;
      }
    }

    public Brush Brush {
      get {
        return ships[0].brush;
      }
      set {
        foreach (Ship ship in ships) {
          ship.brush = value;
        }
      }
    }

    public void Update(int x1, int y1, int x2, int y2) {
      if (exploding) {
        if (!exploded) {
          ships[explodeIndex].explode = true;
          squariusForm.explosions.Add(
              new Explosion(squariusForm, ships[explodeIndex]));
          if (explodeIndex == 0) {
            exploded = true;
          } else {
            explodeIndex--;
          }
        }
      } else {
        double vx = x2 - x1;
        double vy = y2 - y1;
        vx *= scale;
        vy *= scale;
        double x = x1;
        double y = y1;
        for (int i = 0; i < ships.Length; i++) {
          x += vx;
          y += vy;
          ships[i].x = (int)x;
          ships[i].y = (int)y;
        }
      }
    }

    public void Explode() {
      if (exploding) {
        return;
      }
      exploding = true;
      explodeIndex = ships.Length - 1;
    }

    public void Vanish() {
      foreach (TailShip ship in ships) {
        ship.explode = true;
      }
    }
  }

  public class Drip : Ship {

    public double Y;
    public double DY;
    public int startY;

    public Drip(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
       this.Y = this.startY = this.y = y - 32;
      this.x = x;
      this.brush = Brushes.Aqua;
      DY = 1;
    }

    public override bool Move() {
      DY += 0.75;
      Y += DY;
      y = (int)Y;
      if (y > 480) {
        Y = y = (int)startY;
        DY = 1;
      }
      return x + width < squariusForm.viewportX;      
    }

    public override bool Shot() {
      return false;
    }
  }

  public class TailSeekerShip : Ship {

    public double ax;
    public double ay;
    public double vx;
    public double vy;
    public Tail tail;
    public int hits;
    public int flashing;
    public bool alive;
    public Brush defaultBrush;

    public TailSeekerShip(SquariusForm squariusForm, int x, int y, bool yieldsPowerUp)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      defaultBrush = this.brush = yieldsPowerUp ? Brushes.Orange : Brushes.Aquamarine;
      this.yieldsPowerUp = yieldsPowerUp;
      tail = new Tail(squariusForm, x, y, width, height, 2, 16, brush, this);
      alive = true;
    }

    public override bool Move() {
      if (alive) {
        Ship playerShip = squariusForm.playerShip;
        double ax = playerShip.x - x;
        double ay = playerShip.y - y;
        double mag = Math.Sqrt(ax * ax + ay * ay);
        if (mag > 0) {
          double scale = 0.3 / mag;
          ax *= scale;
          ay *= scale;
          vx += ax;
          vy += ay;
          vx *= 0.98;
          vy *= 0.98;
          x = (int)(x + vx);
          y = (int)(y + vy);
        }
        if (flashing > 0) {
          flashing--;
          if (flashing == 0) {
            tail.Brush = defaultBrush;
            brush = defaultBrush;
          }
        }
      }
      tail.Follow(x, y);
      return tail.exploded;
    }

    public override bool Shot() {
      if (flashing == 0 && alive) {
        flashing = 2;
        tail.Brush = brush = Brushes.White;
        hits++;
        if (hits == 8) {
          tail.Explode();
          alive = false;
        }
      }
      return tail.exploded;
    }
  }
}