/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_6() {
      scrollSpeed = 2;
      wallBrush = Brushes.White;
      Text = "SQUARIUS\u00B2 - Stage 7";
    }

    public void ProcessCell_6(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          WaveShip.CreateHelix(this, x, y, 64, false);
          break;
        case 'B':
          WaveShip.CreateHelix(this, x, y, 64, true);
          break;
        case 'C':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'D':
          ships.Add(new HorizontalShip(this, x, y, true, true));
          break;
        case 'E':
          ships.Add(new ScrollSpeedUp(this, x, y));
          break;
        case 'Y':
          ships.Add(new HighSpeedBoss(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class ScrollSpeedUp : Ship {

    public enum State { Slowing, Paused, Reverse, Forward, Halt }

    public double speed = 2;
    public State state = State.Slowing;
    public int pauseDelay = 10;
    public double vx = -1;
    public bool selfDestruct;
    
    public ScrollSpeedUp(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      if (squariusForm.scrollSpeed != 2) {
        selfDestruct = true;
      }
    }

    public override bool Move() {
      if (selfDestruct) {
        return true;
      }
      switch (state) {
        case State.Slowing:
          speed -= 0.01;
          if (speed <= 0.5) {
            speed = 0;
            state = State.Reverse;
          }
          squariusForm.scrollSpeed = (int)speed;
          break;
        case State.Paused:
          if (pauseDelay > 0) {
            pauseDelay--;
          } else {
            state = State.Reverse;
          }
          break;
        case State.Reverse:
          vx += 0.05;
          speed += vx;
          if (vx >= 0) {
            state = State.Forward;
          }
          squariusForm.scrollSpeed = (int)speed;
          break;
        case State.Forward:
          if (speed < 16) {
            squariusForm.scrollSpeed = (int)speed;
            vx += 0.1;
            speed += vx;
          } else {
            squariusForm.scrollSpeed = 16;
            state = State.Halt;
            List<Bullet> bullets = squariusForm.bullets;
            for (int i = bullets.Count - 1; i >= 0; i--) {
              Bullet bullet = bullets[i];
              if (bullet.vx < 34) {
                bullets.RemoveAt(i);
              }
            }
            int inc = 16 + squariusForm.viewportX 
                - ((squariusForm.viewportX | 0x1F) + 1);
            squariusForm.viewportX += inc;
            squariusForm.playerShip.x += inc;
          }          
          break;
        case State.Halt:
          return true;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Beam : Ship {

    public bool vanish;

    public Beam(SquariusForm squariusForm) : base(squariusForm) {
    }

    public override bool Move() {
      return vanish;
    }

    public override bool Shot() {
      return vanish;
    }
  }

  public class HighSpeeder : Ship {

    public HighSpeeder(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.width = 128;
      brush = Brushes.White;
    }

    public override bool Move() {
      x -= 16;
      if (x + width < squariusForm.viewportX) {
        x = squariusForm.viewportX + 640;
        y = (squariusForm.random.Next(2) == 0) ? 32 : 416;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class HighSpeedBoss : Ship {

    public enum State { Entry, ShootingFromBehind, Advancing, MovingUp, 
        ExtendingBeams, MovingDown, MovingCenter, Squeeze, BeamRelease,
        ShootingFront, Halt };

    public Brush defaultBrush;
    public State state = State.Entry;
    public int shots = 10;
    public int shootDelay = 30;
    public Beam upperBeam;
    public Beam lowerBeam;
    public int flashing;
    public int hits;

    public HighSpeedBoss(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      this.x = x - 896;
      this.y = y;
      this.width = this.height = 224;
      defaultBrush = this.brush = Brushes.Tan;
      squariusForm.ships.Add(new HighSpeeder(squariusForm,
          squariusForm.viewportX + 736,
          (squariusForm.random.Next(2) == 0) ? 32 : 416));
    }

    public override bool Move() {
      switch (state) {
        case State.Entry:
          x++;
          if (x == 32 + squariusForm.viewportX) {
            state = State.ShootingFromBehind;
          }
          break;
        case State.ShootingFromBehind:
          if (shootDelay > 0) {
            shootDelay--;
          } else {
            shootDelay = 30;
            squariusForm.bullets.Add(new Bullet(squariusForm, this));
            shots--;
            if (shots == 0) {
              state = State.Advancing;
            }
          }
          break;
        case State.Advancing:
          x++;
          if (x == 384 + squariusForm.viewportX) {
            state = State.MovingUp;
          }
          break;
        case State.MovingUp:
          y--;
          if (y == 64) {
            state = State.ExtendingBeams;
            upperBeam = new Beam(squariusForm);
            upperBeam.brush = Brushes.MediumSlateBlue;
            upperBeam.x = x - 16;
            upperBeam.y = y + 8;
            upperBeam.width = 16;
            upperBeam.height = 16;
            squariusForm.ships.Add(upperBeam);
            lowerBeam = new Beam(squariusForm);
            lowerBeam.brush = Brushes.MediumSlateBlue;
            lowerBeam.x = x - 16;
            lowerBeam.y = y + height - 25;
            lowerBeam.width = 16;
            lowerBeam.height = 17;
            squariusForm.ships.Add(lowerBeam);
          }
          break;
        case State.ExtendingBeams:
          if (upperBeam.x > squariusForm.viewportX) {
            upperBeam.x -= 16;
            upperBeam.width += 16;
            lowerBeam.x -= 16;
            lowerBeam.width += 16;
          } else {
            state = State.MovingDown;
          }
          break;
        case State.MovingDown:
          if (y < 192) {
            y++;
            upperBeam.y++;
            lowerBeam.y++;
          } else {
            state = State.MovingCenter;
          }
          break;
        case State.MovingCenter:
          if (y > 128) {
            y--;
            upperBeam.y--;
            lowerBeam.y--;
          } else {
            state = State.Squeeze;
          }
          break;
        case State.Squeeze:
          if (lowerBeam.y - upperBeam.y > 75) {
            lowerBeam.y--;
            upperBeam.y++;
          } else {
            state = State.BeamRelease;
          }
          break;
        case State.BeamRelease:
          upperBeam.x -= 16;
          lowerBeam.x -= 16;
          if (upperBeam.x + upperBeam.width < squariusForm.viewportX) {
            upperBeam.vanish = true;
            lowerBeam.vanish = true;
            state = State.ShootingFront;
            shootDelay = 30;
            shots = 10;
          }
          break;
        case State.ShootingFront:
          if (shootDelay > 0) {
            shootDelay--;
          } else {
            shootDelay = 30;
            squariusForm.bullets.Add(new Bullet(squariusForm, this));
            shots--;
            if (shots == 0) {
              state = State.MovingUp;
            }
          }
          break;
      }
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
        }
      }
      return false;
    }

    public override bool Shot() {
      if (state == State.ShootingFront && flashing == 0) {
        hits++;
        if (hits == 100) {
          defaultBrush = Brushes.Magenta;
        }
        if (hits == 150) {
          state = State.Halt;
          squariusForm.requestAdvanceStage = true;
          return true;
        }
        flashing = 2;
        brush = Brushes.White;
      }
      return false;
    }
  }
}