/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Squarius2 {
  public partial class SquariusForm {

    public void InitStage_7() {
      scrollSpeed = 2;
      wallBrush = Brushes.MediumPurple;
      Text = "SQUARIUS\u00B2 - Stage 8";
    }

    public void ProcessCell_7(char cell, int x, int y) {
      switch (cell) {
        case 'A':
          ships.Add(new AppearingSeekerShip(this, x - 150, y, false));
          break;
        case 'B':
          ships.Add(new AppearingSeekerShip(this, x - 550, y, false));
          break;
        case 'C':
          ships.Add(new AppearingSeekerShip(this, x - 50, y, true));
          break;
        case 'D':
          ships.Add(new Fan(this, x, y, 0.02));
          break;
        case 'E':
          ships.Add(new Fan(this, x, y, -0.03));
          break;
        case 'F':
          ships.Add(new Creeper(this, x, y, true, false));
          break;
        case 'G':
          ships.Add(new Creeper(this, x, y, false, false));
          break;
        case 'H':
          ships.Add(new HorizontalShip(this, x, y, false, true));
          break;
        case 'I':
          WaveShip.CreateHelix(this, x, y, 64, false);
          break;
        case 'J':
          RightAngleShip.CreateFleet(this, x, y, false);
          break;
        case 'K':
          ships.Add(new HorizontalShip(this, x, y, false, false));
          break;
        case 'L':
          ships.Add(new Elevator(this, x, y, 1, 0.02));
          break;
        case 'M':
          ships.Add(new Elevator(this, x, y, 3, 0.02));
          break;
        case 'N':
          ships.Add(new WallGun(this, x, y, false));
          break;
        case 'O':
          ships.Add(new Crusher(this, x, y, 160, 0, 128, 120, true));
          break;
        case 'P':
          ships.Add(new Crusher(this, x, y, 320, 32, 128, 80, false));
          break;
        case 'Q':
          ships.Add(new Crusher(this, x, y, 192, 32, 288, 120, false));
          break;
        case 'R':
          ships.Add(new Crusher(this, x, y, 64, 0, 320, 80, true));
          break;
        case 'S':
          ships.Add(new Crusher(this, x, y, 416, 0, 176, 140, true));
          break;
        case 'T':
          ships.Add(new Crusher(this, x, y, 416, 32, 176, 200, false));
          break;
        case 'U':
          ships.Add(new LargeWallGun(this, x, y, false));
          break;
        case 'a':
          ships.Add(new Robot(this, x, y));
          break;
        case 'b':
          ships.Add(new ElectricField(this, x, y, 32, 160, 0));
          break;
        case 'c':
          ElectricField.CreatePropagatingField(this, x, y, 32, 160);
          break;
        case 'd':
          ships.Add(new Bouncer(this, x - 672, y, 0.5));
          break;
        case 'e':
          ships.Add(new Bouncer(this, x - 672, y, -0.5));
          break;
        case 'f':
          ships.Add(new Sperm(this, x, y));
          break;
        case 'V':
          ships.Add(new VShip(this, x, y));
          break;
        case 'W':
          ships.Add(new BattleShip(this, x, y));
          break;
        case 'Y':
          ships.Add(new RubiksCube(this, x, y));
          break;
        case 'Z':
          scrollSpeed = 0;
          break;
        case '$':
          MakeCheckPoint();
          break;
      }
    }
  }

  public class SpermTail : Ship {

    public Ribbon ribbon;
    public double seekX;
    public double seekY;
    public double angle;
    public Sperm sperm;
    public int hits;
    public int flashing;

    public SpermTail(SquariusForm squariusForm, int x, int y, Sperm sperm)
        : base(squariusForm) {
      this.sperm = sperm;
      seekX = this.x = x;
      seekY = this.y = y;
      this.brush = Brushes.GreenYellow;
      ribbon = new Ribbon(squariusForm, 32, 32, 24, brush, this, true,
          0, 4.0 * Math.PI, 32);
      ribbon.Update(x, y, (int)seekX, (int)seekY);
    }

    public override bool Move() {
      Ship playerShip = squariusForm.playerShip;
      double vx = playerShip.x - seekX;
      double vy = playerShip.y - seekY;
      double scale = 3 / Math.Sqrt(vx * vx + vy * vy);
      vx *= scale;
      vy *= scale;
      seekX += vx;
      seekY += vy;
      angle += 0.06;
      if (!ribbon.exploding && x + width < squariusForm.viewportX) {
        ribbon.Explode();        
      }
      ribbon.angleOffset += 0.1;
      ribbon.Update(sperm.x + 32, sperm.y + 32, (int)seekX, (int)seekY);
      if (ribbon.exploded) {
        sperm.TailExploded();
      }
      if (flashing > 0) {
        flashing--;
      }
      return ribbon.exploded;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        sperm.TailShot();
        hits++;
        if (hits == 10) {
          ribbon.Explode();
        }
      }
      return false;
    }
  }

  public class Sperm : Ship {

    public int flashing;
    public Brush defaultBrush;
    public double angle;
    public int Y;
    public SpermTail tail;
    public bool dead;

    public Sperm(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      this.x = x;
      Y = this.y = y;
      width = height = 96;
      squariusForm.ships.Add(
          tail = new SpermTail(squariusForm, x + 32, y + 32, this));
      brush = defaultBrush = Brushes.GreenYellow;
    }

    public override bool Move() {
      angle += 0.1;
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
          tail.brush = defaultBrush;
          tail.ribbon.Brush = defaultBrush;
        }
      }
      y = (int)(Y + 16 * Math.Sin(angle));
      dead = x + width < squariusForm.viewportX;
      return dead;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        brush = Brushes.White;
        tail.brush = Brushes.White;
        tail.ribbon.Brush = Brushes.White;
      }
      return false;
    }

    public void TailExploded() {
      if (!dead) {
        squariusForm.ships.Add(
            tail = new SpermTail(squariusForm, x + 32, y + 32, this));
      }
    }

    public void TailShot() {
      Shot();
    }
  }

  public class Bouncer : Ship {

    public double X;
    public double Y;
    public double VY;
    public double G;
    public double groundY;

    public Bouncer(SquariusForm squariusForm, int x, int y,
        double g) : base(squariusForm) {
      X = this.x = x;
      groundY = Y = this.y = y;      
      G = g;
      if (G > 0) {
        VY = -13;
      } else {
        VY = 13;
      }
      brush = Brushes.Tan;
    }

    public override bool Move() {

      X += 5;
      Y += VY;
      VY += G;
      if (G > 0) {
        if (VY > 0 && Y >= groundY) {
          VY = -13;
          Y = groundY;
        }
      } else {
        if (VY < 0 && Y <= groundY) {
          VY = 13;
          Y = groundY;
        }
      }
      x = (int)X;
      y = (int)Y;

      return x > squariusForm.viewportX + 650;
    }

    public override bool Shot() {
      return true;
    }
  }

  public class ElectricField : Ship {

    public int delay;
    public int WIDTH;
    public int HEIGHT;

    public static void CreatePropagatingField(SquariusForm squariusForm, int x, int y,
        int width, int height) {
      for(int i = 0, k = x, d = 0; i < 5; i++, k += 64, d += 12) {
        squariusForm.ships.Add(new ElectricField(
            squariusForm, k, y, 32, 160, d));
      }
    }

    public ElectricField(SquariusForm squariusForm, int x, int y,
        int width, int height, int delay) : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.delay = delay;
      this.width = 0;
      this.height = 0;
      WIDTH = width;
      HEIGHT = height;
    }

    public override bool Move() {
      if (delay > 0) {
        delay--;
      } else {
        delay = 30;
        if (width == 0) {
          width = WIDTH;
          height = HEIGHT;
        } else {
          width = height = 0;
        }
      }
      brush = (brush == Brushes.White) ? Brushes.MediumSlateBlue : Brushes.White;
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Robot : Ship {

    public Brush defaultBrush;
    public Ribbon upperLeft;
    public Ribbon upperRight;
    public Ribbon lowerLeft;
    public Ribbon lowerRight;
    public Point2D upperLeftPoint;
    public Point2D upperRightPoint;
    public Point2D lowerLeftPoint;
    public Point2D lowerRightPoint;
    public TailShip upperLeftFoot;
    public TailShip lowerLeftFoot;
    public TailShip upperRightFoot;
    public TailShip lowerRightFoot;
    public int dx = 3;
    public double yAngle;
    public int Y;
    public int flashing;
    public bool exploding;
    public int shootDelay;

    public double legAngleInc = Math.PI / 60.0;

    public double lowerRightAngle;
    public int lowerRightCount;
    public int lowerRightDX;

    public double lowerLeftAngle;
    public int lowerLeftCount;
    public int lowerLeftDX;

    public double upperRightAngle;
    public int upperRightCount;
    public int upperRightDX;

    public double upperLeftAngle;
    public int upperLeftCount;
    public int upperLeftDX;    

    public Robot(SquariusForm squariusForm, int x, int y) : base(squariusForm) {
      defaultBrush = brush = Brushes.Blue;
      x -= 900;
      this.x = x;
      Y = this.y = y;
      width = height = 96;
      upperLeft = new Ribbon(squariusForm, 32, 32, 20, brush, this, true,
          0, Math.PI, -96);
      upperRight = new Ribbon(squariusForm, 32, 32, 20, brush, this, true,
          0, Math.PI, 96);
      lowerLeft = new Ribbon(squariusForm, 32, 32, 20, brush, this, true,
          0, Math.PI, 96);
      lowerRight = new Ribbon(squariusForm, 32, 32, 20, brush, this, true,
          0, Math.PI, -96);
      upperLeftPoint = new Point2D(x - 160, 32);
      upperRightPoint = new Point2D(x + 64, 32);
      lowerLeftPoint = new Point2D(x - 64, 416);
      lowerRightPoint = new Point2D(x + 160, 416);
      upperLeftFoot = new TailShip(squariusForm, this);
      upperLeftFoot.width = upperLeftFoot.height = 48;
      upperRightFoot = new TailShip(squariusForm, this);
      upperRightFoot.width = upperRightFoot.height = 48;
      lowerLeftFoot = new TailShip(squariusForm, this);
      lowerLeftFoot.width = lowerLeftFoot.height = 48;
      lowerRightFoot = new TailShip(squariusForm, this);
      lowerRightFoot.width = lowerRightFoot.height = 48;
      upperLeftFoot.brush = upperRightFoot.brush = lowerLeftFoot.brush
          = lowerRightFoot.brush = brush;
      squariusForm.ships.Add(upperLeftFoot);
      squariusForm.ships.Add(upperRightFoot);
      squariusForm.ships.Add(lowerLeftFoot);
      squariusForm.ships.Add(lowerRightFoot);
      Move();
    }

    public override bool Move() {
      if (!exploding && dx == 0 && x - squariusForm.viewportX < 320) {
        exploding = true;
        lowerRightFoot.explode = true;
        upperRightFoot.explode = true;
        lowerLeftFoot.explode = true;
        upperLeftFoot.explode = true;
        squariusForm.explosions.Add(new Explosion(squariusForm, lowerRightFoot));
        squariusForm.explosions.Add(new Explosion(squariusForm, upperRightFoot));
        squariusForm.explosions.Add(new Explosion(squariusForm, lowerLeftFoot));
        squariusForm.explosions.Add(new Explosion(squariusForm, upperLeftFoot));
        lowerRight.Explode();
        lowerLeft.Explode();
        upperRight.Explode();
        upperLeft.Explode();
      } else if (exploding && upperLeft.exploded && upperRight.exploded
          && lowerLeft.exploded && lowerRight.exploded) {
        squariusForm.explosions.Add(new Explosion(squariusForm, this));
        return true;
      } else if (!exploding && dx == 0) {
        if (shootDelay > 0) {
          shootDelay--;
        } else {
          shootDelay = 40;
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
        }
      }

      if (x > squariusForm.viewportX + 900) {
        dx = 0;
        lowerRightCount = 1;
        lowerLeftCount = 1;
        upperRightCount = 1;
        upperLeftCount = 1;
      }
      if (lowerRightCount > 0) {
        lowerRightCount--;
        lowerRightPoint.x += lowerRightDX;
        lowerRightAngle += legAngleInc;
        lowerRightPoint.y = (int)(416 - 96 * Math.Sin(lowerRightAngle));
        if (lowerRightCount == 0) {
          lowerRightPoint.y = 416;
        }
      } else if (x + 160 - lowerRightPoint.x > 64) {
        lowerRightCount = 60;
        lowerRightDX = (x + 400 - lowerRightPoint.x) / 50;
        lowerRightAngle = 0;
      }
      if (lowerLeftCount > 0) {
        lowerLeftCount--;
        lowerLeftPoint.x += lowerLeftDX;
        lowerLeftAngle += legAngleInc;
        lowerLeftPoint.y = (int)(416 - 96 * Math.Sin(lowerLeftAngle));
        if (lowerLeftCount == 0) {
          lowerLeftPoint.y = 416;
        }
      } else if ((x - 160 - lowerLeftPoint.x) > 64) {
        lowerLeftCount = 60;
        lowerLeftDX = (x + 80 - lowerLeftPoint.x) / 50;
        lowerLeftAngle = 0;
      }
      if (upperRightCount > 0) {
        upperRightCount--;
        upperRightPoint.x += upperRightDX;
        upperRightAngle += legAngleInc;
        upperRightPoint.y = (int)(32 + 96 * Math.Sin(upperRightAngle));
        if (upperRightCount == 0) {
          upperRightPoint.y = 32;
        }
      } else if ((x + 160 - upperRightPoint.x) > 64) {
        upperRightCount = 60;
        upperRightDX = (x + 400 - upperRightPoint.x) / 50;
        upperRightAngle = 0;
      }
      if (upperLeftCount > 0) {
        upperLeftCount--;
        upperLeftPoint.x += upperLeftDX;
        upperLeftAngle += legAngleInc;
        upperLeftPoint.y = (int)(32 + 96 * Math.Sin(upperLeftAngle));
        if (upperLeftCount == 0) {
          upperLeftPoint.y = 32;
        }
      } else if ((x - 160 - upperLeftPoint.x) > 64) {
        upperLeftCount = 60;
        upperLeftDX = (x + 80 - upperLeftPoint.x) / 50;
        upperLeftAngle = 0;
      }

      x += dx;
      yAngle += 0.075;
      y = (int)(Y + 8 * Math.Sin(yAngle));
      upperLeft.Update(x + 32, y + 32, upperLeftPoint.x, upperLeftPoint.y);
      upperRight.Update(x + 32, y + 32, upperRightPoint.x, upperRightPoint.y);
      lowerLeft.Update(x + 32, y + 32, lowerLeftPoint.x, lowerLeftPoint.y);
      lowerRight.Update(x + 32, y + 32, lowerRightPoint.x, lowerRightPoint.y);
      upperLeftFoot.x = upperLeftPoint.x - 16;
      upperLeftFoot.y = upperLeftPoint.y;
      upperRightFoot.x = upperRightPoint.x;
      upperRightFoot.y = upperRightPoint.y;
      lowerLeftFoot.x = lowerLeftPoint.x - 16;
      lowerLeftFoot.y = lowerLeftPoint.y - 16;
      lowerRightFoot.x = lowerRightPoint.x;
      lowerRightFoot.y = lowerRightPoint.y - 16;
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
          upperLeft.Brush = defaultBrush;
          lowerLeft.Brush = defaultBrush;
          upperRight.Brush = defaultBrush;
          lowerRight.Brush = defaultBrush;
          upperLeftFoot.brush = defaultBrush;
          upperRightFoot.brush = defaultBrush;
          lowerLeftFoot.brush = defaultBrush;
          lowerRightFoot.brush = defaultBrush;
        }
      }
      return false;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        brush = Brushes.White;
        upperLeft.Brush = Brushes.White;
        lowerLeft.Brush = Brushes.White;
        upperRight.Brush = Brushes.White;
        lowerRight.Brush = Brushes.White;
        upperLeftFoot.brush = Brushes.White;
        upperRightFoot.brush = Brushes.White;
        lowerLeftFoot.brush = Brushes.White;
        lowerRightFoot.brush = Brushes.White;
      }
      return false;
    }
  }

  public class Crusher : Ship {

    public int targetHeight;
    public bool expandsDown;
    public int delay;

    public Crusher(SquariusForm squariusForm, int x, int y, int width, int height,
        int targetHeight, int delay, bool expandsDown) : base(squariusForm) {
      this.x = x;
      this.y = y;
      this.width = width;
      this.height = height;
      this.targetHeight = targetHeight;
      this.delay = delay;
      this.expandsDown = expandsDown;
      this.brush = squariusForm.wallBrush;
    }

    public override bool Move() {
      if (delay > 0) {
        delay--;
      } else if (height < targetHeight) {
        height += 4;
        if (!expandsDown) {
          y -= 4;
        }
      }
      return x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Elevator : Ship {

    public double angle;
    public double angleInc;
    public TailShip bottom;

    public Elevator(SquariusForm squariusForm, int x, int y, double angle,
        double angleInc) : base(squariusForm) {
      this.x = x;
      this.y = 32;
      this.angle = angle;
      this.angleInc = angleInc;
      bottom = new TailShip(squariusForm, this);
      bottom.x = x;
      bottom.width = width = 256;
      bottom.height = 0;
      bottom.brush = brush = Brushes.BlanchedAlmond;
      squariusForm.ships.Add(bottom);
      Move();
    }

    public override bool Move() {
      angle += angleInc;
      int Y = (int)(176 + 144 * Math.Sin(angle));
      height = Y - 32;
      bottom.y = Y + 128;
      bottom.height = 448 - bottom.y;
      if (x + width < squariusForm.viewportX) {
        bottom.explode = true;
        return true;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class Fan : Ship {

    public double angle0 = 0;
    public double angle1 = 2.0 * Math.PI / 3.0;
    public double angle2 = 4.0 * Math.PI / 3.0;
    public Line blade0;
    public Line blade1;
    public Line blade2;
    public double angleInc;

    public Fan(SquariusForm squariusForm, int x, int y, double angleInc) 
        : base(squariusForm) {
      this.x = x + 192;
      this.y = y;
      this.angleInc = angleInc;
      brush = squariusForm.wallBrush;      
      blade0 = new Line(squariusForm, 32, 32, 16, brush, this, false);
      blade1 = new Line(squariusForm, 32, 32, 16, brush, this, false);
      blade2 = new Line(squariusForm, 32, 32, 16, brush, this, false);
      Move();
    }

    public override bool Move() {
      angle0 += angleInc;
      angle1 += angleInc;
      angle2 += angleInc;

      blade0.Update(x, y, 
          (int)(x + 192 * Math.Cos(angle0)), 
          (int)(y + 192 * Math.Sin(angle0)));
      blade1.Update(x, y,
          (int)(x + 192 * Math.Cos(angle1)),
          (int)(y + 192 * Math.Sin(angle1)));
      blade2.Update(x, y,
          (int)(x + 192 * Math.Cos(angle2)),
          (int)(y + 192 * Math.Sin(angle2)));
      if (x + 225 < squariusForm.viewportX) {
        blade0.Vanish();
        blade1.Vanish();
        blade2.Vanish();
        return true;
      }
      return false;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class AppearingSeekerShip : Ship {

    public double ax;
    public double ay;
    public double vx;
    public double vy;
    public bool appearing;

    public AppearingSeekerShip(SquariusForm squariusForm, int x, int y, 
        bool yieldsPowerUp) : base(squariusForm) {
      this.x = x;
      this.y = y;
      width = height = 2;
      appearing = true;
      this.brush = yieldsPowerUp ? Brushes.Orange : Brushes.Aquamarine;
      this.yieldsPowerUp = yieldsPowerUp;
    }

    public override bool Move() {
      if (appearing) {
        x--;
        y--;
        width += 2;
        height += 2;
        if (width == 32) {
          appearing = false;
        }
      } else {
        Ship playerShip = squariusForm.playerShip;
        double ax = playerShip.x - x;
        double ay = playerShip.y - y;
        double mag = Math.Sqrt(ax * ax + ay * ay);
        if (mag > 0) {
          double scale = 0.3 / mag;
          ax *= scale;
          ay *= scale;
          vx += ax;
          vy += ay;
          vx *= 0.98;
          vy *= 0.98;
          x = (int)(x + vx);
          y = (int)(y + vy);
        } else {
          return true;
        }
      }
      return false;
    }

    public override bool Shot() {
      if (appearing) {
        return false;
      }
      return true;
    }
  }

  public class VShip : Ship {

    public enum State { Entry, Expand, Rotate, Shooting, Exploding }

    public State state = State.Entry;
    public Brush defaultBrush;
    public int entryDelay = 64;
    public Line upperLine;
    public Line lowerLine;
    public int X;
    public int Y;
    public int radius;
    public double upperAngle = Math.PI;
    public double lowerAngle = Math.PI;
    public int rotateSteps = 80;
    public double upperAngleInc = Math.PI / 160.0;
    public double lowerAngleInc = -Math.PI / 160.0;
    public double moveAngle;
    public int shootingDelay = 10;
    public int flashing;
    public int hits;

    public VShip(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      defaultBrush = brush = Brushes.DarkOrange;
      X = this.x = x;
      Y = this.y = y;
      width = height = 96;
      upperLine = new Line(squariusForm, 32, 32, 16, brush, this, false);
      lowerLine = new Line(squariusForm, 32, 32, 16, brush, this, false);
    }

    public override bool Move() {
      switch (state) {
        case State.Entry:
          if (entryDelay > 0) {
            entryDelay--;
          } else {
            state = State.Expand;
          }
          break;
        case State.Expand:
          radius += 4;
          upperLine.Update(X + 32, Y + 32, X + 32 - radius, Y + 32);
          lowerLine.Update(X + 32, Y + 32, X + 32 - radius, Y + 32);
          if (radius == 240) {
            state = State.Rotate;
          }
          break;
        case State.Rotate:
          upperAngle += upperAngleInc;
          lowerAngle += lowerAngleInc;
          upperLine.Update(X + 32, Y + 32,
              (int)(X + 32 + 240 * Math.Cos(upperAngle)),
              (int)(Y + 32 + 240 * Math.Sin(upperAngle)));
          lowerLine.Update(X + 32, Y + 32,
              (int)(X + 32 + 240 * Math.Cos(lowerAngle)),
              (int)(Y + 32 + 240 * Math.Sin(lowerAngle)));
          rotateSteps--;
          if (rotateSteps == 0) {
            state = State.Shooting;
          }
          break;
        case State.Shooting:
          moveAngle += 0.05;
          x = (int)(X + 24 * Math.Sin(moveAngle));
          y = (int)(Y + 32 * Math.Sin(1.7 * moveAngle));
          upperLine.Update(x + 32, y + 32, x + 32, y - 208);
          lowerLine.Update(x + 32, y + 32, x + 32, y + 272);
          if (shootingDelay > 0) {
            shootingDelay--;
          } else {
            shootingDelay = 120;
            int safeY = squariusForm.random.Next(416);
            int minSafeY = Y + safeY - 208;
            int maxSafeY = minSafeY + 64;
            for (int i = Y - 176; i < Y + 304; i += 16) {
              if (i < minSafeY || i > maxSafeY) {
                for (int j = 0, k = x + 32; j < 3; j++, k += 8) {
                  squariusForm.bullets.Add(new Bullet(squariusForm, BulletType.Enemy,
                      k, i, -5, 0));
                }
              }
            }
          }
          if (flashing > 0) {
            flashing--;
            if (flashing == 0) {
              brush = defaultBrush;
              upperLine.Brush = defaultBrush;
              lowerLine.Brush = defaultBrush;
            }
          }
          break;
        case State.Exploding:
          upperLine.Update(x + 32, y + 32, x + 32, y - 208);
          lowerLine.Update(x + 32, y + 32, x + 32, y + 272);
          if (upperLine.exploded && lowerLine.exploded) {
            squariusForm.scrollSpeed = 2;
            return true;
          }
          break;
      }
      return false;
    }

    public override bool Shot() {
      if (state == State.Shooting && flashing == 0) {
        flashing = 2;
        brush = Brushes.White;
        upperLine.Brush = Brushes.White;
        lowerLine.Brush = Brushes.White;
        hits++;
        if (hits == 250) {
          defaultBrush = brush = Brushes.Magenta;
        } else if (hits == 300) {
          brush = Brushes.Magenta;
          upperLine.Brush = brush;
          lowerLine.Brush = brush;
          state = State.Exploding;
          upperLine.Explode();
          lowerLine.Explode();
        }
      }
      return false;
    }
  }

  public class BattleShipGun : Ship {

    public Brush defaultBrush;
    public int xOffset;
    public int yOffset;
    public BattleShip battleShip;
    public int shootDelay;
    public int flashing;
    public int hits;

    public BattleShipGun(SquariusForm squarisForm,
        BattleShip battleShip, int xOffset, int yOffset) : base(squarisForm) {
      this.battleShip = battleShip;
      this.xOffset = xOffset;
      this.yOffset = yOffset;
      x = xOffset + battleShip.x;
      y = yOffset + battleShip.y;
      width = height = 64;
      defaultBrush = brush = Brushes.Olive;
      shootDelay = squarisForm.random.Next(120);
    }

    public override bool Move() {
      x = xOffset + battleShip.x;
      y = yOffset + battleShip.y;
      if (shootDelay > 0) {
        shootDelay--;
      } else {
        shootDelay = 120;
        squariusForm.bullets.Add(new Bullet(squariusForm, this));
      }
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
        }
      }
      return false;
    }

    public override bool Shot() {
      if (flashing == 0) {
        flashing = 2;
        brush = Brushes.White;
        hits++;
        if (hits == 25) {
          battleShip.GunDestroyed();
          return true;
        }
      }
      return false;
    }
  }

  public class BattleShip : Ship {

    public Brush defaultBrush;
    public double yAngle;
    public double xAngle;
    public double yAngle2;
    public int Y;
    public int X;
    public int gunsRemaining = 7;
    public bool vunerable;
    public bool blowingUp;
    public int flashing;
    public int shootDelay;
    public int hits;
    public int explosionDelay;

    public BattleShip(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      X = this.x = x;
      Y = this.y = y;
      width = 576;
      height = 192;
      defaultBrush = brush = Brushes.Olive;

      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, -64, 64));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 64, -64));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 256, -64));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 448, -64));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 64, 192));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 256, 192));
      squariusForm.ships.Add(new BattleShipGun(squariusForm, this, 448, 192));
    }

    public override bool Move() {
      if (!blowingUp) {
        yAngle += 0.01;
        xAngle += 0.005;
        yAngle2 += 0.003;
      }
      y = (int)(Y - 16 * Math.Sin(yAngle) + 128 * Math.Sin(yAngle2));
      x = (int)(X - 320 * Math.Sin(xAngle));
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          brush = defaultBrush;
        }
      }
      if (blowingUp) {
        Y += 2;
        if (y > 480) {
          squariusForm.scrollSpeed = 2;
          return true;
        }
        if (explosionDelay > 0) {
          explosionDelay--;
        } else {
          squariusForm.explosions.Add(new Explosion(squariusForm, 
              x + squariusForm.random.Next(width),
              y + squariusForm.random.Next(height)));
          explosionDelay = 2;
        }
      } else if (vunerable) {
        if (shootDelay > 0) {
          shootDelay--;
        } else {
          squariusForm.bullets.Add(new Bullet(squariusForm, this));
          shootDelay = squariusForm.random.Next(50);
        }
      }
      return false;
    }

    public override bool Shot() {
      if (vunerable && flashing == 0 && !blowingUp) {
        flashing = 2;
        brush = Brushes.White;
        hits++;
        if (hits > 50) {
          blowingUp = true;
        }
      }
      return false;
    }

    public void GunDestroyed() {
      gunsRemaining--;
      if (gunsRemaining == 0) {
        defaultBrush = brush = Brushes.Magenta;
        vunerable = true;
        shootDelay = 40;
      }
    }
  }

  public class FaceShip : Ship {

    public bool explode;
    public bool vanish;
    public Ship head;
    public Brush backup = Brushes.White;

    public void Flashing(bool on) {
      if (on) {
        backup = brush;
        brush = Brushes.White;
      } else {
        brush = backup;
      }
    }

    public FaceShip(SquariusForm squariusForm, Ship head)
        : base(squariusForm) {
      this.head = head;
    }

    public override bool Move() {
      return explode || vanish;
    }

    public override bool Shot() {
      head.Shot();
      return explode;
    }
  }

  public class Drop : Ship {

    public double Y;
    public double DY;
    public int startY;

    public Drop(SquariusForm squariusForm, int x, int y, Brush[] colors)
        : base(squariusForm) {
      this.Y = this.startY = this.y = y;
      this.x = x;
      brush = colors[squariusForm.random.Next(colors.Length)];
      DY = 1;
      this.width = this.height = 63;
    }

    public override bool Move() {
      DY += 0.75;
      Y += DY;
      y = (int)Y;
      return y > 480 || x + width < squariusForm.viewportX;
    }

    public override bool Shot() {
      return false;
    }
  }

  public class RubiksCube : Ship {

    public enum State { Entry, Shooting, MovingUp, Dropping, Returning, Bouncing }

    public enum Sides { Up = 0, Down = 1, Left = 2, Right = 3, Front = 4, Back = 5 }
    public Brush[,,] faces = new Brush[6, 3, 3];
    public Brush[] colors = { Brushes.White, Brushes.Blue, Brushes.Red,
        Brushes.Orange, Brushes.Yellow, Brushes.Green };
    public FaceShip[,] frontSquares = new FaceShip[3, 3];
    public FaceShip[] extraSquares = new FaceShip[3];
    public Point2D[,,] points = new Point2D[3,3,2];
    public Point2D[, ,] rotatedPoints = new Point2D[3, 3, 2];
    public bool rotating;
    public bool horizontalRotation;
    public bool clockwiseRotation;
    public double rotationAngle;
    public double rotationAngleInc;
    public int rotationCount;
    public int rotationIndex;
    public int rotationDelay;
    public double startMoveAngle = 19.8; // y = 145
    public double moveAngle;
    public int X;
    public int flashing;
    public State state = State.Entry;
    public int entryDelay = 120;
    public int lastX;
    public int lastY;
    public int shotCount;
    public double droppingAngle;
    public int droppingDelay;
    public int dropCount;
    public int vx;
    public int vy;
    public int bounceCount;
    public int hits;
    public int droppingCenter;

    public RubiksCube(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      this.X = this.x = x;
      this.y = 145;
      this.width = 0;
      this.height = 0;
      this.brush = Brushes.Black;

      moveAngle = startMoveAngle;
      x = (int)(X + 32 * Math.Cos(moveAngle) + 16 * Math.Sin(moveAngle * .7));
      y = (int)(144 + 48 * Math.Cos(moveAngle * 1.2 + 0.3)
          + 96 * Math.Sin(moveAngle * .3 + .2));

      InitFaces();
      InitPoints();
      for (int i = 0; i < 3; i++) {
        FaceShip ship = new FaceShip(squariusForm, this);
        ship.width = 0;
        ship.height = 63;
        extraSquares[i] = ship;
      }
      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          FaceShip ship = new FaceShip(squariusForm, this);
          ship.height = ship.width = 63;
          ship.backup = ship.brush = faces[(int)Sides.Front, i, j];
          ship.x = x + (j << 6);
          ship.y = y + (i << 6);
          squariusForm.ships.Add(ship);
          frontSquares[i, j] = ship;
        }
      }

      InitRotation(squariusForm.random.Next(2) == 0, 
          squariusForm.random.Next(2) == 0,
          squariusForm.random.Next(3));
    }

    public void InitRotation(bool horizontal, bool clockwise, int index) {
      rotating = true;
      horizontalRotation = horizontal;
      clockwiseRotation = clockwise;
      rotationIndex = index;
      rotationAngle = 0;
      rotationCount = 40;
      rotationAngleInc = Math.PI / (2.0 * rotationCount);
      if (!clockwise) {
        rotationAngleInc = -rotationAngleInc;
      }

      if (horizontal) {
        if (clockwise) {
          extraSquares[0].backup = extraSquares[0].brush 
              = faces[(int)Sides.Right, index, 0];
          extraSquares[1].backup = extraSquares[1].brush 
              = faces[(int)Sides.Right, index, 1];
          extraSquares[2].backup = extraSquares[2].brush 
              = faces[(int)Sides.Right, index, 2];
        } else {
          extraSquares[0].backup = extraSquares[0].brush 
              = faces[(int)Sides.Left, index, 0];
          extraSquares[1].backup = extraSquares[1].brush 
              = faces[(int)Sides.Left, index, 1];
          extraSquares[2].backup = extraSquares[2].brush 
              = faces[(int)Sides.Left, index, 2];
        }
        for (int i = 0; i < 3; i++) {
          extraSquares[i].height = 63;
          extraSquares[i].width = 0;
          extraSquares[i].x = 0;
          extraSquares[i].y = y + (index << 6);
        }
      } else {
        if (clockwise) {
          extraSquares[0].backup = extraSquares[0].brush 
              = faces[(int)Sides.Up, 0, index];
          extraSquares[1].backup = extraSquares[1].brush 
              = faces[(int)Sides.Up, 1, index];
          extraSquares[2].backup = extraSquares[2].brush 
              = faces[(int)Sides.Up, 2, index];
        } else {
          extraSquares[0].backup = extraSquares[0].brush 
              = faces[(int)Sides.Down, 0, index];
          extraSquares[1].backup = extraSquares[1].brush 
              = faces[(int)Sides.Down, 1, index];
          extraSquares[2].backup = extraSquares[2].brush 
              = faces[(int)Sides.Down, 2, index];
        }
        for (int i = 0; i < 3; i++) {
          extraSquares[i].width = 63;
          extraSquares[i].height = 0;
          extraSquares[i].x = x + (index << 6);
          extraSquares[i].y = 0;
        }
      }

      for (int i = 0; i < 3; i++) {
        extraSquares[i].vanish = false;        
        squariusForm.ships.Add(extraSquares[i]);
      }
    }

    public void InitPoints() {
      
      // front
      points[0, 0, 0] = new Point2D(-96, 96);
      points[0, 0, 1] = new Point2D(-96 + 63, 96);
      points[0, 1, 0] = new Point2D(-96 + 64, 96);
      points[0, 1, 1] = new Point2D(-96 + 64 + 63, 96);
      points[0, 2, 0] = new Point2D(-96 + 128, 96);
      points[0, 2, 1] = new Point2D(-96 + 128 + 63, 96);

      // left
      points[1, 0, 0] = new Point2D(-96, - 96);
      points[1, 0, 1] = new Point2D(-96, -96 + 63);
      points[1, 1, 0] = new Point2D(-96, -96 + 64);
      points[1, 1, 1] = new Point2D(-96, -96 + 64 + 63);
      points[1, 2, 0] = new Point2D(-96, -96 + 128);
      points[1, 2, 1] = new Point2D(-96, -96 + 128 + 63);

      // right
      points[2, 0, 0] = new Point2D(96, -96);
      points[2, 0, 1] = new Point2D(96, -96 + 63);
      points[2, 1, 0] = new Point2D(96, -96 + 64);
      points[2, 1, 1] = new Point2D(96, -96 + 64 + 63);
      points[2, 2, 0] = new Point2D(96, -96 + 128);
      points[2, 2, 1] = new Point2D(96, -96 + 128 + 63);

      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          for (int k = 0; k < 2; k++) {
            rotatedPoints[i, j, k] = new Point2D();
          }
        }
      }
    }

    public void InitFaces() {
      for (int face = 0; face < 6; face++) {
        for (int i = 0; i < 3; i++) {
          for (int j = 0; j < 3; j++) {
            faces[face, i, j] = colors[face];
          }
        }
      }
    }

    public void TurnMagenta() {
      for (int face = 0; face < 6; face++) {
        for (int i = 0; i < 3; i++) {
          for (int j = 0; j < 3; j++) {
            faces[face, i, j] = Brushes.Magenta;
          }
        }
      }
      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          FaceShip ship = frontSquares[i, j];
          ship.brush = ship.backup = Brushes.Magenta;
        }
      }
      for (int i = 0; i < extraSquares.Length; i++) {
        extraSquares[i].backup = extraSquares[i].brush = Brushes.Magenta;
      }
    }

    public void RotateFaces(bool horizontal, bool clockwise, int index) {
      if (horizontal) {
        if (clockwise) {
          HorizontalClockwiseRotate(index);
        } else {
          HorizontalCounterclockwiseRotate(index);
        }
        for (int i = 0; i < 3; i++) {
          FaceShip ship = frontSquares[index, i];
          ship.backup = ship.brush = faces[(int)Sides.Front, index, i];
        }
      } else {
        if (clockwise) {
          VerticleClockwiseRotate(index);
        } else {
          VerticleCounterclockwiseRotate(index);
        }
        for (int i = 0; i < 3; i++) {
          FaceShip ship = frontSquares[i, index];
          ship.backup = ship.brush = faces[(int)Sides.Front, i, index];
        }
      }
    }

    public void VerticleClockwiseRotate(int column) {
      SwapColumns(column, Sides.Front, Sides.Up);
      SwapColumns(column, Sides.Up, Sides.Back);
      SwapColumns(column, Sides.Back, Sides.Down);
    }

    public void VerticleCounterclockwiseRotate(int column) {
      SwapColumns(column, Sides.Front, Sides.Down);
      SwapColumns(column, Sides.Down, Sides.Back);
      SwapColumns(column, Sides.Back, Sides.Up);
    }

    public void HorizontalClockwiseRotate(int row) {
      SwapRows(row, Sides.Front, Sides.Right);
      SwapRows(row, Sides.Right, Sides.Back);
      SwapRows(row, Sides.Back, Sides.Left);
    }

    public void HorizontalCounterclockwiseRotate(int row) {
      SwapRows(row, Sides.Front, Sides.Left);
      SwapRows(row, Sides.Left, Sides.Back);
      SwapRows(row, Sides.Back, Sides.Right);
    }

    public void SwapRows(int row, Sides side1, Sides side2) {
      for (int i = 0; i < 3; i++) {
        Brush temp = faces[(int)side2, row, i];
        faces[(int)side2, row, i] = faces[(int)side1, row, i];
        faces[(int)side1, row, i] = temp;
      }
    }

    public void SwapColumns(int column, Sides side1, Sides side2) {
      for (int i = 0; i < 3; i++) {
        Brush temp = faces[(int)side2, i, column];
        faces[(int)side2, i, column] = faces[(int)side1, i, column];
        faces[(int)side1, i, column] = temp;
      }
    }

    public override bool Move() {
      bool retVal = false;
      switch (state) {
        case State.Entry:
          if (entryDelay > 0) {
            entryDelay--;
          } else {
            state = State.Shooting;
            shotCount = 8 + squariusForm.random.Next(9);
            moveAngle = startMoveAngle;
          }
          break;
        case State.Shooting:
          retVal = ShootPlayer();
          break;
        case State.MovingUp:
          if (y > 0) {
            y -= 6;
          }
          if (Math.Abs(squariusForm.viewportX + 224 - x) > 6) {
            if (x > squariusForm.viewportX + 224) {
              x -= 6;
            } else if (x < squariusForm.viewportX + 224) {
              x += 6;
            }
          } else if (y <= 0) {
            y = 0;
            x = squariusForm.viewportX + 224;
            droppingAngle = 0;
            dropCount = 8;
            state = State.Dropping;
            droppingCenter = squariusForm.viewportX + 224;
          }
          UpdateFrontSquares();
          break;
        case State.Dropping:
          x = droppingCenter
              + (int)(75 * Math.Sin(2.3 * droppingAngle)
                  + 149 * Math.Sin(droppingAngle));
          if (x < squariusForm.viewportX) {
            x = (squariusForm.viewportX << 1) - x;
          } else if (x > squariusForm.viewportX + 448) {
            x = ((squariusForm.viewportX + 448) << 1) - x;
          }
          int targetDroppingCenter = squariusForm.playerShip.x - 80;
          if (droppingCenter > targetDroppingCenter) {
            droppingCenter -= 2;
          } else if (droppingCenter < targetDroppingCenter) {
            droppingCenter += 2;
          }
          droppingAngle += 0.05;
          if (droppingDelay > 0) {
            droppingDelay--;
          } else if (x - 31 <= squariusForm.playerShip.x 
              && x + 191 >= squariusForm.playerShip.x){
            droppingDelay = squariusForm.random.Next(40) + 60;
            int dropX = x + (squariusForm.random.Next(3) << 6);
            int dropY = y + (squariusForm.random.Next(3) << 6);
            squariusForm.ships.Add(new Drop(squariusForm, dropX, dropY, colors));
            if (dropCount > 0) {
              dropCount--;
            } else {
              state = State.Bouncing;
              bounceCount = 400;
              vy = 2;
              vx = 6;
            }
          }
          UpdateFrontSquares();
          break;
        case State.Bouncing:
          x += vx;
          y += vy;
          if (x <= squariusForm.viewportX) {
            squariusForm.bullets.Add(new Bullet(squariusForm, frontSquares[1, 1]));
            x = squariusForm.viewportX;
            vx = squariusForm.random.Next(3) + 6;
          } else if (x >= squariusForm.viewportX + 448) {
            squariusForm.bullets.Add(new Bullet(squariusForm, frontSquares[1, 1]));
            x = squariusForm.viewportX + 448;
            vx = -(squariusForm.random.Next(3) + 6);
          }
          if (y <= 0) {
            squariusForm.bullets.Add(new Bullet(squariusForm, frontSquares[1, 1]));
            y = 0;
            vy = squariusForm.random.Next(2) + 3;
          } else if (y >= 288) {
            squariusForm.bullets.Add(new Bullet(squariusForm, frontSquares[1, 1]));
            y = 288;
            vy = -(squariusForm.random.Next(2) + 3);
          }
          if (bounceCount > 0) {
            bounceCount--;
          } else {
            state = State.Returning;
          }
          UpdateFrontSquares();
          break;
        case State.Returning:
          if (Math.Abs(x - lastX) > 6) {
            if (x > lastX) {
              x -= 6;
            } else if (x < lastX) {
              x += 6;
            }
          } else if (Math.Abs(y - lastY) > 6) {
            x = lastX;
            if (y > lastY) {
              y -= 6;
            } else if (y < lastY) {
              y += 6;
            }
          } else {
            y = lastY;
            shotCount = 8 + squariusForm.random.Next(9);
            state = State.Shooting;
          }
          UpdateFrontSquares();
          break;
      }
      return retVal;
    }

    public void UpdateFrontSquares() {
      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          FaceShip ship = frontSquares[i, j];
          ship.x = x + (j << 6);
          ship.y = y + (i << 6);
        }
      }
    }

    public bool ShootPlayer() {

      x = (int)(X + 32 * Math.Cos(moveAngle) + 16 * Math.Sin(moveAngle * .7));
      y = (int)(144 + 48 * Math.Cos(moveAngle * 1.2 + 0.3)
          + 96 * Math.Sin(moveAngle * .3 + .2));
      moveAngle += 0.1;
      if (flashing > 0) {
        flashing--;
        if (flashing == 0) {
          for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
              frontSquares[i, j].Flashing(false);
            }
            extraSquares[i].Flashing(false);
          }
        }
      }

      UpdateFrontSquares();

      if (rotating) {
        double sin = Math.Sin(rotationAngle);
        double cos = Math.Cos(rotationAngle);
        for (int i = 0; i < 3; i++) {
          for (int j = 0; j < 3; j++) {
            for (int k = 0; k < 2; k++) {
              Point2D point = points[i, j, k];
              Point2D rotatedPoint = rotatedPoints[i, j, k];

              rotatedPoint.x = (int)(cos * point.x - sin * point.y);
              rotatedPoint.y = (int)(sin * point.x + cos * point.y);
            }
          }
        }

        if (horizontalRotation) {
          for (int col = 0; col < 3; col++) {
            FaceShip face = frontSquares[rotationIndex, col];
            int x1 = rotatedPoints[0, col, 0].x;
            int x2 = rotatedPoints[0, col, 1].x;
            face.x = x + 96 + Math.Min(x1, x2);
            face.y = y + (rotationIndex << 6);
            face.width = Math.Max(0, Math.Abs(x2 - x1));
          }
          if (clockwiseRotation) {
            for (int col = 0; col < 3; col++) {
              FaceShip face = extraSquares[col];
              int x1 = rotatedPoints[2, 2 - col, 0].x;
              int x2 = rotatedPoints[2, 2 - col, 1].x;
              face.x = x + 96 + Math.Min(x1, x2);
              face.y = y + (rotationIndex << 6);
              face.width = Math.Max(0, Math.Abs(x2 - x1));
            }
          } else {
            for (int col = 0; col < 3; col++) {
              FaceShip face = extraSquares[col];
              int x1 = rotatedPoints[1, col, 0].x;
              int x2 = rotatedPoints[1, col, 1].x;
              face.x = x + 96 + Math.Min(x1, x2);
              face.y = y + (rotationIndex << 6);
              face.width = Math.Max(0, Math.Abs(x2 - x1));
            }
          }
        } else {
          for (int col = 0; col < 3; col++) {
            FaceShip face = frontSquares[col, rotationIndex];
            int x1 = rotatedPoints[0, col, 0].x;
            int x2 = rotatedPoints[0, col, 1].x;
            face.x = x + (rotationIndex << 6);
            face.y = y + 96 + Math.Min(x1, x2);
            face.height = Math.Max(0, Math.Abs(x2 - x1));
          }
          if (clockwiseRotation) {
            for (int col = 0; col < 3; col++) {
              FaceShip face = extraSquares[col];
              int x1 = rotatedPoints[2, 2 - col, 0].x;
              int x2 = rotatedPoints[2, 2 - col, 1].x;
              face.x = x + (rotationIndex << 6);
              face.y = y + 96 + Math.Min(x1, x2);
              face.height = Math.Max(0, Math.Abs(x2 - x1));              
            }
          } else {
            for (int col = 0; col < 3; col++) {
              FaceShip face = extraSquares[col];
              int x1 = rotatedPoints[1, col, 0].x;
              int x2 = rotatedPoints[1, col, 1].x;
              face.x = x + (rotationIndex << 6);
              face.y = y + 96 + Math.Min(x1, x2);
              face.height = Math.Max(0, Math.Abs(x2 - x1));
            }
          }
        }

        rotationAngle += rotationAngleInc;
        rotationCount--;
        if (rotationCount == 0) {
          rotating = false;
          rotationDelay = 5;
          RotateFaces(horizontalRotation, clockwiseRotation, rotationIndex);
          if (horizontalRotation) {
            for (int i = 0; i < 3; i++) {
              extraSquares[i].vanish = true;
              FaceShip ship = frontSquares[rotationIndex, i];
              ship.x = x + (i << 6);
              ship.y = y + (rotationIndex << 6);
              ship.width = ship.height = 63;
            }
          } else {
            for (int i = 0; i < 3; i++) {
              extraSquares[i].vanish = true;
              FaceShip ship = frontSquares[i, rotationIndex];
              ship.x = x + (rotationIndex << 6);
              ship.y = y + (i << 6);
              ship.width = ship.height = 63;
            }
          }

          int shootY = y + squariusForm.random.Next(192);
          squariusForm.bullets.Add(new Bullet(squariusForm, frontSquares[1, 1]));
          for (int i = 0, j = x; i < 10; i++, j += 8) {
            squariusForm.bullets.Add(new Bullet(squariusForm,
                BulletType.Enemy, j, shootY, -8, 0));            
          }
          if (shotCount > 0) {
            shotCount--;
          } else {
            state = State.MovingUp;
            lastX = x;
            lastY = y;
            flashing = 0;
            for (int i = 0; i < 3; i++) {
              for (int j = 0; j < 3; j++) {
                frontSquares[i, j].Flashing(false);
              }
              extraSquares[i].Flashing(false);
            }
          }
        }
      } else {
        rotationDelay--;
        if (rotationDelay == 0) {
          InitRotation(squariusForm.random.Next(2) == 0,
              squariusForm.random.Next(2) == 0,
              squariusForm.random.Next(3));
        }
      }

      return false;
    }

    public override bool Shot() {
      if (flashing == 0 && state == State.Shooting) {
        for (int i = 0; i < 3; i++) {
          for (int j = 0; j < 3; j++) {
            frontSquares[i, j].Flashing(true);
          }
          extraSquares[i].Flashing(true);
        }
        flashing = 2;
        hits++;
        if (hits == 300) {
          TurnMagenta();
        } else if (hits == 375) {
          squariusForm.requestAdvanceStage = true;
        }
      }
      return false;
    }
  }
}